//------------------------------------------------------------------------------
// Purpose: Search for symbols (rather than for documents)
// Date   : 19.Sep.2009
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Windows.Forms;

namespace Fusym
{
    public class StringAndCount
    {
        public string s;
        public int count;
    }

    public class IntegerValue
    {
        public int value;
    }

    public class Betweenness
    {
        private Dictionary<Node, IntegerValue> m_Transits = new Dictionary<Node, IntegerValue>();
        private Random m_Rnd = new Random();



        ///---------------------------------------------------------------------
        /// <summary>
        /// Create a clone of an associative network but omit one node.
        /// </summary>
        ///---------------------------------------------------------------------
        private SemanticNetwork CloneClusterManagerWithException(SemanticNetwork cm, string exception)
        {
            SemanticNetwork result = new SemanticNetwork(cm.TheEncoding);
            List<Node> nodes = cm.NodeList;

            for (int i = 0; i < nodes.Count; i++)
            {
                Node n = nodes[i];

                string s1 = n.ToString();
                if (s1 == exception)
                    continue;


                Node n1 = result.AddSymbol(s1);

                List<Node> neighbours = n.Neighbours;
                for (int j = 0; j < neighbours.Count; j++)
                {
                    Node m = neighbours[j];
                    string s2 = m.ToString();

                    if (s2 == exception)
                        continue;

                    Node n2 = result.AddSymbol(s2);

                    result.AddPair(n1, n2);
                }

            }

            return result;
        }









        public string RunTestCycles(SemanticNetwork cm, int maxTransits)
        {
            string result = null;
            List<Node> nodes = cm.NodeList;
            List<Node> path  = new List<Node>();
            int countOfNodes = nodes.Count;
            m_Transits.Clear();

            while (result == null)
            {
                // Choose two random nodes
                Node startNode = nodes[m_Rnd.Next(0,countOfNodes)];
                Node endNode   = nodes[m_Rnd.Next(0,countOfNodes)];

                bool b = cm.FindPath(startNode.ToString(), endNode.ToString(), path);
                if (b == false)
                {
                    b = false;
                    continue;
                }

                // Increment a transitcount for each node on the path
                for (int i = 0; i < path.Count; i++)
                {
                    Node n = path[i];
                    IntegerValue value;
                    if (m_Transits.TryGetValue(n, out value))
                        value.value++;
                    else
                    {
                        value = new IntegerValue();
                        m_Transits.Add(n, value);
                    }

                    if (value.value >= maxTransits)
                    {
                        result = n.ToString();
                        break;
                    }
                }

            }

            return result;
        }



        public void RobustDistance1(SemanticNetwork cm, string symbol1, string symbol2)
        {
            double d = RobustDistance(cm, symbol1, symbol2);
            MessageBox.Show(symbol1 + " " + symbol2 + " " + d.ToString());
        }




        public string Test(SemanticNetwork cm, int maxTransits)
        {
            RobustDistance1(cm,"wolke"  , "zum");
            RobustDistance1(cm,"wolke", "gruseln");
            RobustDistance1(cm,"zum"  , "gruseln");
            return "";




            string result = "";

            for (int i = 0; i < 100; i++)
            {
                string symbol = RunTestCycles(cm, maxTransits);
                cm = CloneClusterManagerWithException(cm, symbol);
                result += symbol + System.Environment.NewLine;
            }

            return result;
        }





        ///---------------------------------------------------------------------
        /// <summary>
        /// Retrieve a list of connected components. Each is
        /// represented by a single symbol that is a member
        /// of the component.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<StringAndCount> ConnectedComponents(SemanticNetwork cm, string symbol)
        {
            List<StringAndCount> result = new List<StringAndCount>();


            if (cm.GetNode(symbol) == null)
               return result;


            cm.MarkAllNodes(0);

            while (symbol != null)
            {
                int i = MarkConnectedComponent(cm, symbol);

                StringAndCount sc = new StringAndCount();
                sc.s = symbol;
                sc.count = i;
                result.Add(sc);

                symbol = FindFirstUnmarked(cm);
            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Finds the first node that is not marked.
        /// </summary>
        ///---------------------------------------------------------------------
        private string FindFirstUnmarked(SemanticNetwork cm)
        {
            string result = null;

            List<Node> nodes = cm.NodeList;
            for (int i = 0; i < nodes.Count; i++)
            {
                Node n = nodes[i];
                if (n.Mark == 0)
                {
                    result = n.ToString();
                    break;
                }
            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Mark all nodes that are connected with symbol.
        /// </summary>
        ///---------------------------------------------------------------------
        public int MarkConnectedComponent(SemanticNetwork cm, string symbol)
        {
            int result = 0;
            Stack<Node> stack = new Stack<Node>();

            Node n = cm.GetNode(symbol);


            stack.Push(n);

            while (stack.Count != 0)
            {
                n = stack.Pop();
                n.Mark = 1;
                result++;

                List<Node> neighbours = n.Neighbours;
                for (int i = 0; i < neighbours.Count; i++)
                {
                    Node m = neighbours[i];
                    if (m.Mark == 0)
                        stack.Push(m);
                }
            }

            return result;
        }






        public double RobustDistance(SemanticNetwork cm, string symbol1, string symbol2)
        {
            const int cSeven = 7;

            int measurements = 0;
            int sum = 0;


            while (measurements++ <= cSeven)
            {
                Node n1 = cm.GetNode(symbol1);
                Node n2 = cm.GetNode(symbol2);
                List<Node> path = new List<Node>();

                bool b = cm.FindPath(symbol1, symbol2, path);

                if (!b ||(path.Count < 3))
                    break;

                sum += path.Count;

                int max = int.MinValue;
                int maxI = -1;

                for (int i = 1; i < path.Count - 1; i++)
                {
                    Node n = path[i];
                    if (n.Neighbours.Count > max)
                    {
                        max = n.Neighbours.Count;
                        maxI = i;
                    }
                }

                string symbol = path[maxI].ToString();
                cm = CloneClusterManagerWithException(cm, symbol);
            }


            if (measurements > 0)
                return sum / (double)measurements;
            else
                return cSeven;

        }



    }
}
