//------------------------------------------------------------------------------
// Purpose: A BlospExpression is a treelike data structure of strings.
// Date   : 29. Mar. 2008
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using Symps;

namespace Fusym
{
    public class BlospExpression : List<BlospExpressionSymbol>, IDocument, IDocumentList
    {
        private List<string> m_Symbols;
        private int  m_CurrentSymbol;
        private int  m_CurrentDocument;
        private bool m_Deleted = false;


        protected static bool m_SearchConceptNames = false;
        protected static bool m_SearchSlotNames = false;
        protected static bool m_SearchSlotValues = false;


        public bool Deleted
        {
            get { return m_Deleted; }
            set { 
                  m_Deleted = value;
                  this.Clear();
                  if (m_Symbols != null)
                     m_Symbols.Clear();
                }
        }

        public static bool SearchConceptNames
        {
            get { return m_SearchConceptNames; }
            set { m_SearchConceptNames = value; }
        }

        public static bool SearchSlotNames
        {
            get { return m_SearchSlotNames; }
            set { m_SearchSlotNames = value; }
        }

        public static bool SearchSlotValues
        {
            get { return m_SearchSlotValues; }
            set { m_SearchSlotValues = value; }
        }

        public override string ToString()
        {
            return ToString(0);
        }

        public void Delete()
        {
            Deleted = true;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Check if one of the strings in the line matches with name
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public bool HasName(string name)
        {
            name = name.ToUpper();

            foreach (BlospExpressionSymbol bs in this)
                if (bs.Name.ToUpper() == name)
                    return true;

            return false;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Find a line in the children of the first BlospExpressionSymbol with 
        /// the first BlospExpressionSymbols's name = slotName and return that line.
        /// </summary>
        ///---------------------------------------------------------------------
        public BlospExpression FindSlotLine(string slotName)
        {
            if (this.Count <= 0)
               return null;

            slotName = slotName.ToUpper();

            for (int i = 0; i < this[0].Children.Count; i++)
            {
                BlospExpression current = this[0].Children[i];
                if (current.Count > 0)
                {
                   string name = current[0].Name.ToUpper();
                   if (name == slotName)
                      return current;
                }
            }

            return null;
        }



        public int SlotAsInteger(string slotName,  int defaultValue)
        {
            int result;

            BlospExpression be = FindSlotLine(slotName);
            if (
                (be != null) &&
                (be.Count == 2)
               )
            {
                string value = be[1].Name;
                result = Int32.Parse(value);
            }
            else
                result = defaultValue;

            return result;
        }


        public int SlotAsInteger(string slotName)
        {
            return SlotAsInteger(slotName, 0);
        }


        public double SlotAsFloat(string slotName, double defaultValue)
        {
            double result;

            BlospExpression be = FindSlotLine(slotName);
            if (
                (be != null) &&
                (be.Count == 2)
               )
            {
                string value = be[1].Name;
                result = Double.Parse(value);
            }
            else
                result = defaultValue;

            return result;
        }


        public double SlotAsFloat(string slotName)
        {
            return SlotAsFloat(slotName, 0);
        }






        public bool SlotAsBool(string slotName, bool defaultValue)
        {
            bool result;

            BlospExpression be = FindSlotLine(slotName);
            if (
                (be != null) &&
                (be.Count == 2)
               )
            {
                string value = be[1].Name;
                
                if (value.ToLower() == "true")
                   result = true;
                else
                   result = false;                
            }
            else
                result = defaultValue;

            return result;
        }


        public bool SlotAsBool(string slotName)
        {
            return SlotAsBool(slotName, false);
        }









        public string SlotAsString(string slotName, string defaultValue)
        {
            string result;

            BlospExpression be = FindSlotLine(slotName);
            if (
                (be != null) &&
                (be.Count > 0)
               )
            {
                result = be[0].RestOfTheLine;
            }
            else
                result = defaultValue;

            return result;
        }


        public string SlotAsString(string slotName)
        {
            return SlotAsString(slotName, "");
        }



        public List<string> SlotAsMemo(string slotName, List<string> defaultValue)
        {
            List<string> result = new List<string>();

            BlospExpression be = FindSlotLine(slotName);
            if (
                (be != null)
               )
            {
                foreach (BlospExpression c in be[0].Children)
                {
                    result.Add(c[0].RestOfTheLine);
                }
            }
            else
            {
                if (defaultValue != null)
                   result.AddRange(defaultValue);
            }

            return result;
        }

        public List<string> SlotAsMemo(string slotName)
        {
            return SlotAsMemo(slotName, null);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Returns a text representation. It can span over severals lines.
        /// An indentation can be passed.
        /// </summary>
        ///---------------------------------------------------------------------
        public string ToString(int indentation)
        {
            if (this.Count <= 0)
                return "";


            List<int> relativeOffsets = new List<int>();
            int pos = 0;
            foreach (BlospExpressionSymbol bs in this)
            {
                relativeOffsets.Add(pos);
                pos += bs.Name.Length + 1;
            }


            string result = new string(' ', indentation) + this[0].Name;

            // Every first symbol in a line must have RestOfTheLine set.
            if (this.Count > 1)
               result += " " + this[0].RestOfTheLine;


            for (int i = this.Count - 1; i >= 0; i--)
            {
                BlospExpressionSymbol be = this[i];

                for (int j = 0; j < be.Children.Count; j++)
                {
                    BlospExpression bl = be.Children[j];
                    int childoffset = indentation + relativeOffsets[i] + 1;
                    string childAsString = bl.ToString(childoffset);
                    result += System.Environment.NewLine + childAsString;
                }
            }

            return result;
        }


        public void AddName(string name)
        {
            BlospExpressionSymbol symbol = new BlospExpressionSymbol(name, 0);
            this.Add(symbol);
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Set the value of a slot.
        /// </summary>
        ///---------------------------------------------------------------------
        public void SetSlotValue(string slotname, SlotValue value)
        {
            // This line must contain at least one symbol, the name
            if (this.Count <= 0)
               return;

            BlospExpression slotline;

            slotline = FindSlotLine(slotname);

            // A new line for both the slotname and the slotvalue
            if (slotline == null)
            {
                slotline = new BlospExpression();
                slotline.AddName(slotname);
                // The new slot line is child of the first symbol in the line
                this[0].AddChild(slotline);
            }
            else
            {
                // Remove all following BlospExpressionSymbols
                while (slotline.Count > 1)
                    slotline.RemoveAt(slotline.Count - 1);
            }

            BlospExpression symbolsForValue = value.ToBlospExpression();
            foreach (BlospExpressionSymbol bs in symbolsForValue)
                slotline.Add(bs);


            slotline[0].RestOfTheLine = value.TheLine();
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="slotname"></param>
        /// <param name="value"></param>
        ///---------------------------------------------------------------------
        public void SetSlotValue(string slotname, string value)
        {
            SlotValueString sv = new SlotValueString(value);
            SetSlotValue(slotname, sv);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Store a slot value of type memo
        /// </summary>
        /// <param name="slotname"></param>
        /// <param name="value"></param>
        ///---------------------------------------------------------------------
        public void SetSlotValue(string slotname, List<string> value)
        {
            // This line must contain at least one symbol, the name
            if (this.Count <= 0)
                return;

            BlospExpression slotline;

            slotline = FindSlotLine(slotname);

            // A new line for the slotname
            if (slotline == null)
            {
                slotline = new BlospExpression();
                slotline.AddName(slotname);
                // The new slot line is child of the first symbol in the line
                this[0].AddChild(slotline);
            }
            else
            {
                // Remove all following BlospExpressionSymbols
                while (slotline.Count > 1)
                    slotline.RemoveAt(slotline.Count - 1);
            }
            
            slotline[0].Children.Clear();

            // In a first step, give each slot a name from a number
            int i = 0;
            foreach (string s in value)
            {
                slotline.SetSlotValue(i.ToString(), s);
                i++;
            }

            // Rename the slot names
            foreach (BlospExpression be in slotline[0].Children)
            {
                be[0].Name = ":";
            }
        }

        public void SetSlotValue(string slotname, int value)
        {
            SlotValueInteger sv = new SlotValueInteger(value);
            SetSlotValue(slotname, sv);
        }

        public void SetSlotValue(string slotname, double value)
        {
            SlotValueDouble sv = new SlotValueDouble(value);
            SetSlotValue(slotname, sv);
        }

        public void SetSlotValue(string slotname, bool value)
        {
            SlotValueBoolean sv = new SlotValueBoolean(value);
            SetSlotValue(slotname, sv);
        }


        public void SetSlotValue(string slotname, BlospExpression value)
        {
            SlotValueBlospExpression sv = new SlotValueBlospExpression(value);
            SetSlotValue(slotname, sv);
        }



        /*
                public ConceptSystemElement AsConceptSystemElement()
                {
                    if (this.Count == 0)
                        return null;

                    //------------------------------------------------------------
                    // If the first symbol has not children or any that is a slot, 
                    // make a concept out of it, otherwise a concept list.
                    //------------------------------------------------------------
                    if (
                         (this[0].Children == null) ||
                         isSuitedForListOfSlots(this[0].Children)
                        )
                        return AsConcept(this);
                    else
                        return AsConceptList(this);
                }




                ///---------------------------------------------------------------------
                /// <summary>
                /// Test whether a List<List<BlospExpressionSymbol>> is suited for a list of slots
                /// </summary>
                /// <param name="list"></param>
                /// <returns></returns>
                ///---------------------------------------------------------------------
                /// 
                public static bool isSuitedForListOfSlots(List<BlospExpression> list)
                {
                    foreach (BlospExpression element in list)
                        if (isSuitedForSlot(element))
                            return true;

                    return false;
                }


                ///---------------------------------------------------------------------
                /// <summary>
                /// 
                /// </summary>
                /// <param name="list"></param>
                /// <returns></returns>
                ///---------------------------------------------------------------------
                public ConceptList AsConceptList(BlospExpression list)
                {
                    if (list.Count < 1)
                        return null;

                    string name = list[0].Name;

                    ConceptList result = new ConceptList(name);

                    if (list[0].Children != null)
                    {
                        foreach (BlospExpression element in list[0].Children)
                        {
                            result.Add(AsConcept(element));
                        }
                    }

                    //--- Unexpected rest of the line after the concept name ---
                    if (list.Count > 1)
                    {
                        BlospExpression rest = Cdr(list);

                        Concept errorlist1 = new Concept("ERRORLIST");
                        ConceptList errorlist2 = AsConceptList(rest);
                        errorlist1.AddConceptList(errorlist2);
                        result.Add(errorlist1);
                    }

                    return result;
                }

                ///---------------------------------------------------------------------
                /// <summary>
                /// 
                /// </summary>
                /// <returns></returns>
                ///---------------------------------------------------------------------
                public Concept AsConcept()
                {
                    Concept result = new Concept(m_Name);

                    if (m_Children == null)
                        return result;

                    foreach (BlospExpression child in m_Children)
                    {
                        if (isSuitedForSlot(child))
                            result.AddSlot(ListAsSlot(child));
                        else
                            result.AddConceptList(AsConceptList(child));
                    }

                    return result;
                }
                 */



        private static BlospExpression Cdr(BlospExpression list)
        {
            if (list == null)
                return null;

            BlospExpression result = new BlospExpression();
            for (int i = 1; i < list.Count; i++)
                result.Add(list[i]);
            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// A line might be a slot if it contains at least two symbols.
        /// </summary>
        /// <param name="line"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public static bool isSuitedForSlot(List<BlospExpressionSymbol> line)
        {
            if (
                (line != null) &&
                (line.Count >= 2) &&
                (line[0].Children.Count == 0)
               )
                return true;
            else
                return false;
        }




        ///---------------------------------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="list"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        /*
        public Concept AsConcept(BlospExpression list)
        {
            if (list.Count < 1)
                return null;

            Concept result = list[0].AsConcept();

            //--- Unexpected rest of the line after the concept name ---
            if (list.Count > 1)
            {
                BlospExpression rest = Cdr(list);

                ConceptList errorlist = AsConceptList(rest);
                result.AddSlot("ERRORLIST", errorlist);
            }

            return result;
        }
         * */



        ///---------------------------------------------------------------------
        /// <summary>
        /// Build up a list that contains all symbols of the BlospExpression.
        /// </summary>
        ///---------------------------------------------------------------------
        private void PrepareSymbols()
        {
            if (m_Symbols == null)
               m_Symbols = new List<string>();

            if (m_SearchConceptNames)
                AddConceptNames();

            if (m_SearchSlotNames)
               AddSlotNames();

            if (m_SearchSlotValues)
               AddSlotValues();
        }



        private void AddConceptNames()
        {
            for (int i = 0; i < this.Count; i++)
                m_Symbols.Add(this[i].Name);
        }

        private void AddSlotNames()
        {
            if (this.Count <= 0)
               return;

            BlospExpressionSymbol blospExpression = this[0];
            foreach (BlospExpression blospExpressionLine in blospExpression.Children)
            {
                if (blospExpressionLine.Count > 0)
                   m_Symbols.Add(blospExpressionLine[0].Name);
            }
        }

        private void AddSlotValues()
        {
            AddSlotValues(this, 0);
        }

        private void AddSlotValues(BlospExpression line, int level)
        {
            //--- All BlospExpressions line ---
            for (int i = 0; i < line.Count; i++)
            {
                BlospExpressionSymbol blospExpression = line[i];

                if (line != this)
                {
                    if (!((i == 0) && (level == 0))) // If it is not a slot name
                       m_Symbols.Add(blospExpression.Name);
                }

                //--- All children lines below the current BlospExpressionSymbol ---
                foreach (BlospExpression blospExpressionLine in blospExpression.Children)
                    AddSlotValues(blospExpressionLine, level + 1);
            }
        }





        public string GetFirstSymbol()
        {
            PrepareSymbols();
            m_CurrentSymbol = -1;
            return GetNextSymbol();
        }

        public string GetNextSymbol()
        {
            m_CurrentSymbol++;
            if ((m_CurrentSymbol >= 0) && (m_CurrentSymbol < m_Symbols.Count))
                return m_Symbols[m_CurrentSymbol];
            else
                return "";
        }


        public IDocument GetFirstDocument()
        {
            m_CurrentDocument = -1;
            return GetNextDocument();
        }

        public IDocument GetNextDocument()
        {
            m_CurrentDocument++;

            if (this.Count <= 0)
               return null;

            if (m_CurrentDocument < this[0].Children.Count)
               return this[0].Children[m_CurrentDocument];
            else
               return null;
        }
    }
}
