//------------------------------------------------------------------------------
// Purpose: Parser for BLOSP (.bsp) files
// Date   : 20. Aug. 2007
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using Symps;



namespace Fusym
{
    public class BlospParser
    {
        private List<string> m_Lines = new List<string>();
        private List<BlospExpression> m_IndentedStrings = new List<BlospExpression>();
        private BlospExpressionSymbol root = new BlospExpressionSymbol("root", -1);

        public BlospParser(List<string> lines)
        {
            m_Lines.AddRange(lines);
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Parse a stringlist.
        /// </summary>
        ///---------------------------------------------------------------------
        public BlospExpression Parse()
        {
            ReadIndentedStrings();
            CombineLinesFromLastToFirst();

            BlospExpression result = new BlospExpression();
            result.Add(root);

            return result;
        }










        ///---------------------------------------------------------------------
        /// <summary>
        /// Split a string into a list of the string parts and indentations
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        private BlospExpression Split(string s)
        {
            BlospExpression result = new BlospExpression();
            string buffer = "";
            int indentation = 0;
            int i;


            i = 0;
            while (i < s.Length)
            {
                if (s[i] == ' ')
                {
                    if (buffer.Length > 0)
                    {
                        result.Add(new BlospExpressionSymbol(buffer, indentation));
                        buffer = "";
                    }
                }
                else
                {
                    if (buffer.Length == 0)
                        indentation = i;

                    buffer += s[i];
                }

                i++;
            }

            if (buffer.Length > 0)
               result.Add(new BlospExpressionSymbol(buffer, indentation));

            TryToGetRestOfTheLineInRawFormat(result, s);
            

            return result;
        }







        private void TryToGetRestOfTheLineInRawFormat(List<BlospExpressionSymbol> result, string s)
        {
            if (result.Count < 2)
               return;

            BlospExpressionSymbol first = result[0];

            int start = first.Indentation + first.Name.Length + 1;
            int end   = s.Length - start;
            first.RestOfTheLine =  s.Substring(start, end);
        }







        ///---------------------------------------------------------------------
        /// <summary>
        /// Translate the stringlist m_Lines into the list of
        /// indented string m_IndentedStrings.
        /// </summary>
        ///---------------------------------------------------------------------
        private void ReadIndentedStrings()
        {
            foreach (string line in m_Lines)
            {
                BlospExpression indentedStrings = Split(line);
                m_IndentedStrings.Add(indentedStrings);
            }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Search for the BlospExpressionSymbol element of the line which is the
        /// parent element of another BlospExpressionSymbol whose indentation is given.
        /// </summary>
        /// <param name="line"></param>
        /// <param name="indentation"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        private BlospExpressionSymbol GetSuperior(List<BlospExpressionSymbol> line, int indentation)
        {
            for (int i = 0; i < line.Count; i++)
            {
                BlospExpressionSymbol s1 = line[i];
                BlospExpressionSymbol s2 = null;

                if (i < line.Count - 1)
                   s2 = line[i + 1];

                if (
                    (indentation > s1.Indentation) &&
                    ((s2 == null) || (indentation <= s2.Indentation))
                   )
                   return s1;
            }

            return null;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Reads a list of lines from the last line back to the bottom.
        /// </summary>
        ///---------------------------------------------------------------------
        private void CombineLinesFromLastToFirst()
        {
            for (int i = m_IndentedStrings.Count - 1; i >= 0; i--)
            {
                BlospExpression line = m_IndentedStrings[i];

                if (line.Count == 0)
                    continue;

                //--- Indentation of the current line ---
                int indentation = line[0].Indentation;

                //--- Search for the parent line. That is one of   ---
                //--- the lines above, with a shorter indentation. ---
                bool found = false;
                for (int j = i - 1; j >= 0; j--)
                {
                    BlospExpression potentialParentLine = m_IndentedStrings[j];

                    //--- No word in this line? ---
                    if (potentialParentLine.Count == 0)
                        continue;

                    //--- If found, make line to one of the children of potentialParentLine ---
                    if (potentialParentLine[0].Indentation < indentation)
                    {
                        found = true;
                        BlospExpressionSymbol parent = GetSuperior(potentialParentLine, indentation);

                        parent.InsertChild(0,line);
                        break;
                    }
                }

                //--- Parent still not found. Then root is defined to be the parent ---
                if (!found)
                    root.InsertChild(0, line);
            }
        }



    }
}
