//------------------------------------------------------------------------------
// Purpose: 
// Date   : 06. Jun. 2010
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;

namespace Fusym
{
    public class Cluster
    {
        public List<Node> Items;
        private double m_LD = -1;

        public double LD
        {
            get
            {
                if (m_LD < 0)
                    m_LD = LevenshteinDistanceInternal();

                return m_LD;
            }
        }

        public int Count
        {
            get
            {
                return Items.Count;
            }
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Constructor.
        /// </summary>
        ///---------------------------------------------------------------------
        public Cluster(List<Node> items)
        {
            Items = new List<Node>(items);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Internal similarity.
        /// </summary>
        ///---------------------------------------------------------------------
        private double LevenshteinDistanceInternal()
        {
            int sum = 0;
            int count = 0;

            for (int i = 0; i < Count; i++)
                for (int j = 0; j < i; j++)
                {
                    sum += LevenshteinDistance.Compute(Items[i].ToString(), Items[j].ToString());
                    count++;
                }

            if (count > 0)
                return sum / (double)count;
            else
                return 0;
        }


        public override string ToString()
        {
            List<Node> alreadyHandled = new List<Node>(Count);

            string result = "";
            for (int i = 0; i < Count; i++)
            {
                Node n = Items[i];

                if (alreadyHandled.Contains(n))
                    continue;

                result += n.ToString() + Environment.NewLine;
                alreadyHandled.Add(n);
            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// It is assumed that within one cluster all symbols
        /// are distinct. 
        /// </summary>
        ///---------------------------------------------------------------------
        public int CountOfCommonSymbols(Cluster other)
        {
            int result = 0;

            List<Node> itemsOther = other.Items;
            int countOther = other.Count;

            for (int i = 0; i < Count; i++)
            {
                Node n = Items[i];
                for (int j = 0; j < countOther; j++)
                    if (n == itemsOther[j])
                        result++;
            }

            return result;
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Returns 1 if there is a common symbol, 0 otherwise.
        /// </summary>
        ///---------------------------------------------------------------------
        public int HasCommonSymbol(Cluster other)
        {
            List<Node> itemsOther = other.Items;
            int countOther = other.Count;

            for (int i = 0; i < Count; i++)
            {
                Node n = Items[i];
                for (int j = 0; j < countOther; j++)
                    if (n == itemsOther[j])
                        return 1;
            }

            return 0;
        }


        public int LevenShteinDistance(Cluster other)
        {
            int sum = 0;

            List<Node> itemsOther = other.Items;
            
            int countThis  = this.Count;
            int countOther = other.Count;

            for (int i = 0; i < countThis; i++)
            {
                Node a = Items[i];
                string a_string = a.ToString();
                for (int j = 0; j < countOther; j++)
                {
                    Node b = itemsOther[j];
                    sum += LevenshteinDistance.Compute(a_string, b.ToString());
                }
            }

            return sum;
        }




        ///---------------------------------------------------------------------
        /// <summary>
        /// Check if this cluster completely contains another one.
        /// </summary>
        ///---------------------------------------------------------------------
        public bool Contains(Cluster other)
        {
            int countThis = Count;

            List<Node> itemsOther = other.Items;
            int countOther = other.Count;

            for (int i = 0; i < countOther; i++)
            {
                Node otherNode = itemsOther[i];
                bool found = false;
                for (int j = 0; j < Count; j++)
                    if (Items[j] == otherNode)
                    {
                        found = true;
                        break;
                    }

                if (!found)
                    return false;
            }

            return true;
        }







        ///---------------------------------------------------------------------
        /// <summary>
        /// Delete all the items.
        /// </summary>
        ///---------------------------------------------------------------------
        public void Clear()
        {
            Items.Clear();
        }
    }
}
