// Joachim Pimiskern, 6.6.2010, Freeware
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace Fusym
{
    public class ClusterList
    {
        public static Encoding encoding = System.Text.Encoding.GetEncoding(1252);
        private SemanticNetwork m_SemanticNetwork = new SemanticNetwork(ClusterList.encoding);
        public List<Cluster> Items;
        Random m_Rnd = new Random();

        public ClusterList()
        {
            Items = new List<Cluster>();
        }

        public ClusterList(int count)
        {
            Items = new List<Cluster>(count);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Load a list of empty-line-separated clusters
        /// from a textfile.
        /// </summary>
        ///---------------------------------------------------------------------
        public void LoadFromFile(string filename)
        {
            List<Node> cluster = new List<Node>();
            Regex regexSpace   = new Regex(@"^\s*$", RegexOptions.Compiled);
            Regex regexNoSpace = new Regex(@"^\s*\S+\s*$", RegexOptions.Compiled);


            StreamReader sr = null;
            try
            {
                sr = new StreamReader(new BufferedStream(new FileStream(filename, FileMode.Open), 10000000), ClusterList.encoding);

                string line = "";


                while ((line = sr.ReadLine()) != null)
                {
                    if (regexNoSpace.IsMatch(line))
                    {
                        Node sym = m_SemanticNetwork.AddSymbol(line.ToUpper().Trim());
                        cluster.Add(sym);
                    }
                    else
                    if (regexSpace.IsMatch(line))
                    {
                        AddCluster(cluster);
                        cluster.Clear();
                    }
                    else
                        line = line; // unknown line type.
                }

                AddCluster(cluster);
                cluster.Clear();
            }
            finally
            {
                sr.Close();
            }
        }

        private void AddCluster(List<Node> cluster)
        {
            if (cluster.Count == 0)
                return;


            Items.Add(new Cluster(cluster));
            m_SemanticNetwork.AddCluster(cluster);
        }


        public int NeighbourValue(int idx)
        {
            int result = 0;

            if (idx > 0)
                result += Items[idx].HasCommonSymbol(Items[idx - 1]);

            if (idx < Items.Count - 1)
                result += Items[idx].HasCommonSymbol(Items[idx + 1]);

            return result;
        }


        public int LongestRun()
        {
            int max = 1;
            int len = 1;
            for (int i = 1; i < Items.Count; i++)
            {
                Cluster c1 = Items[i-1];
                Cluster c2 = Items[i - 0];

                if (c1.HasCommonSymbol(c2) > 0)
                    len++;
                else
                {
                    if (len > max)
                        max = len;
                    len = 1;
                }
            }

            return max;
        }




        public int CountOfRuns()
        {
            int result = 1;
            for (int i = 1; i < Items.Count; i++)
            {
                Cluster c1 = Items[i - 1];
                Cluster c2 = Items[i - 0];

                if (c1.HasCommonSymbol(c2) == 0)
                    result++;
            }

            return result;
        }
        
        
        
        
        
        
        
        
        
        
        
        public void Swap(int idx1, int idx2)
        {
            Cluster h = Items[idx1];
            Items[idx1] = Items[idx2];
            Items[idx2] = h;
        }


        public void Clear()
        {
            Items.Clear();            
        }




        public void SaveToFile(string filename)
        {
            string clusterOld = "";

            StringBuilder sb = new StringBuilder();
            int count = Items.Count;

            for (int i = 0; i < count; i++)
            {
                string clusterNew = Items[i].ToString();

                if (clusterNew != clusterOld)
                {
                    sb.AppendLine(clusterNew);
                    clusterOld = clusterNew;
                }
            }

            SaveToFile(filename, sb.ToString());
        }


        public bool SaveToFile(string filename, string s)
        {
            try
            {
                FileStream fs = new FileStream(filename, FileMode.Create);
                if (!fs.CanWrite)
                {
                    fs.Close();
                    return false;
                }

                StreamWriter writer = new StreamWriter(fs, System.Text.Encoding.GetEncoding(1252));
                writer.WriteLine(s);
                writer.Flush();
                writer.Close();
                fs.Close();
                return true;
            }
            catch
            {
                return false;
            }
        }

        public void Scramble()
        {
            int count = Items.Count;
            
            for (int i = 0; i < count; i++)
            {
                Cluster h = Items[i];
                int idx = m_Rnd.Next(0, count);
                Items[i] = Items[idx];
                Items[idx] = h;
            }
        }




        public int MinimalSemanticDistance(Cluster a, Cluster other)
        {
            int result = int.MaxValue;

            List<Node> path = new List<Node>();
            List<Node> itemsA     = a.Items;
            List<Node> itemsOther = other.Items;
            int countA     = a.Count;
            int countOther = other.Count;

            for (int i = 0; i < countA; i++)
            {
                Node n1 = a.Items[i];
                for (int j = 0; j < countOther; j++)
                {
                    Node n2 = other.Items[j];
                    if (m_SemanticNetwork.FindPath(n1, n2, path))
                    {
                        int d = path.Count;
                        if (d < result)
                            result = d;
                    }
                }
            }

            return result;
        }

        private int CountOfSymbols()
        {
            Dictionary<Node, int> dict = new Dictionary<Node,int>();
            foreach (Cluster c in Items)
                foreach (Node n in c.Items)
                    dict[n] = 0;

            int result = 0;
            foreach (KeyValuePair<Node, int> pair in dict)
                result++;

            return result;
        }

        private  int MaxClusterSize()
        {
            int result = 0;

            foreach (Cluster c in Items)
                result = Math.Max(result, c.Items.Count);

            return result;
        }
    }
}
