//------------------------------------------------------------------------------
// Purpose: An instance of the class Concept represents a proposition.
// Date   : 21.Jul.2007
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using Fusym;


namespace Symps
{
    public class Knowledgebase : IDocumentList
    {
        private List<ConceptSystemElement> m_List = new List<ConceptSystemElement>();
        private int m_Current;
        private Encoding  m_Encoding;

        ///---------------------------------------------------------------------
        /// <summary>
        /// For file access
        /// </summary>
        /// <param name="encoding"></param>
        ///---------------------------------------------------------------------
        public Knowledgebase(Encoding encoding)
        {
            m_Encoding = encoding;
        }


        public void Add(ConceptSystemElement c)
        {
            m_List.Add(c);
        }

        public ConceptSystemElement Get(int idx)
        {
            return m_List[idx];
        }

        public int Count
        {
            get { return m_List.Count; }
        }

        public ConceptSystemElement GetFirst()
        {
            m_Current = -1;
            return GetNext();
        }

        public ConceptSystemElement GetNext()
        {
            m_Current++;
            if (m_Current < m_List.Count)
               return m_List[m_Current];
            else
               return null;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Read concept system elements from a string list.
        /// </summary>
        /// <param name="sl"></param>
        ///---------------------------------------------------------------------
        ///
        /*
        public void ReadConcepts(List<string> sl)
        {
            BlospParser icp = new BlospParser(sl);
            BlospExpression blospExpressionLine = icp.Parse();

            if (blospExpressionLine[0].Children != null)
            {
                foreach (BlospExpression element in blospExpressionLine[0].Children)
                {
                    ConceptSystemElement c = blospExpressionLine.ListAsConceptSystemElement(element);
                    Add(c);
                }
            }
        }
         */


        ///---------------------------------------------------------------------
        /// <summary>
        /// Appends line from a textfile to the internal List m_Records.
        /// If you want to clear the internal representation, use Clear().
        /// </summary>
        /// <param name="filename">The filename of a text file.</param>
        ///---------------------------------------------------------------------
        /*
        public void LoadFromFile(string filename)
        {
            List<string> sl = new List<string>();
            StreamReader sr = null;

            BufferedStream bs = new BufferedStream(new FileStream(filename, FileMode.Open));

            sr = new StreamReader(bs, m_Encoding);
            try
            {
                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    sl.Add(line);
                }

                ReadConcepts(sl);
            }
            finally
            {
                sr.Close();
            }
        }


        */


        ///---------------------------------------------------------------------
        /// <summary>
        /// Saves the file. 
        /// </summary>
        ///---------------------------------------------------------------------
        public void SaveToFile(string filename)
        {
            StringBuilder sb = new StringBuilder("");
            string tempfile = Path.GetDirectoryName(filename) + @"\temp.$$$";

            for (int i = 0; i < m_List.Count; i++)
            {
                ConceptSystemElement element = (ConceptSystemElement)m_List[i];
                if (!element.Deleted)
                {
                    sb.AppendLine(element.ToString());
                    sb.AppendLine();
                }
            }


            FileStream stream = new FileStream(tempfile, FileMode.Create);
            if (!stream.CanWrite)
            {
                stream.Close();
                return;
            }

            StreamWriter writer = new StreamWriter(stream, m_Encoding);
            writer.Write(sb.ToString());
            writer.Flush();
            writer.Close();
            stream.Close();

            File.Delete(filename);
            File.Move(tempfile, filename);
        }





        public IDocument GetFirstDocument()
        {
            m_Current = 0;
            return GetNextDocument();
        }

        public IDocument GetNextDocument()
        {
            if (m_Current < m_List.Count)
               return m_List[m_Current++];
            else
               return null;
        }

    }


















    public abstract class ConceptSystemElement : IDocument
    {
        protected bool m_Deleted;

        protected List<string> listOfAllWords;
        protected int current;
        protected static bool m_SearchConceptNames = false;
        protected static bool m_SearchSlotNames = false;
        protected static bool m_SearchSlotValues = false;

        public static bool SearchConceptNames
        {
            get { return m_SearchConceptNames; }
            set { m_SearchConceptNames = value; }
        }

        public static bool SearchSlotNames
        {
            get { return m_SearchSlotNames; }
            set { m_SearchSlotNames = value; }
        }

        public static bool SearchSlotValues
        {
            get { return m_SearchSlotValues; }
            set { m_SearchSlotValues = value; }
        }



        public bool Deleted
        {
            get { return m_Deleted; }
        }

        public virtual void Delete()
        {
            m_Deleted = true;
        }

        public virtual string GetFirstSymbol()
        {
            return string.Empty;
        }

        public virtual string GetNextSymbol()
        {
            return string.Empty;
        }
    }
























    public class Concept : ConceptSystemElement
    {
        private string m_Name;
        private List<CustomSlot> m_Slots;

        public Concept(string name)
        {
            m_Name = name;
        }

        public string Name
        {
            get { return m_Name; }
        }

        public List<CustomSlot> Slots
        {
            get { return m_Slots; }
        }

        public void AddSlot(string name, object value)
        {
            if (m_Slots == null)
                m_Slots = new List<CustomSlot>();

            m_Slots.Add(new Slot(name,value));
        }

        public void AddSlot(Slot slot)
        {
            if (m_Slots == null)
                m_Slots = new List<CustomSlot>();

            m_Slots.Add(slot);
        }

        public void AddConceptList(ConceptList list)
        {
            if (m_Slots == null)
                m_Slots = new List<CustomSlot>();

            m_Slots.Add(list);
        }

        public string ToStringRestIndented(int indentation)
        {
            if (m_Deleted)
                return string.Empty;

            string prefix = new string(' ', indentation);
            string result = ""; // Evtl. mit StringBuilder zu-tun
            result += m_Name;

            if ((m_Slots != null) && (m_Slots.Count > 0))
            {
                result += System.Environment.NewLine;
                for (int i = 0; i < m_Slots.Count; i++)
                    if (i < m_Slots.Count - 1)
                        result += prefix + m_Slots[i].ToString(indentation + 1) + System.Environment.NewLine;
                    else
                        result += prefix + m_Slots[i].ToString(indentation + 1);
            }

            return result;
        }

        public string ToString(int indentation)
        {
            if (m_Deleted)
               return string.Empty;

           string prefix = new string(' ', indentation);
           return prefix + ToStringRestIndented(indentation);
        }

        public override string ToString()
        {
            return ToString(0);
        }


        public override string GetFirstSymbol()
        {
            if (listOfAllWords == null)
               listOfAllWords = new List<string>(1 + m_Slots.Count * 2);
            else
               listOfAllWords.Clear();

            current = 0;

            if (m_SearchConceptNames)
               listOfAllWords.Add(m_Name);

            if (m_SearchSlotNames)
               for (int i = 0; i < m_Slots.Count; i++)
                   listOfAllWords.Add(m_Slots[i].Name);

           if (m_SearchSlotValues)
               for (int i = 0; i < m_Slots.Count; i++)
                   if ((m_Slots[i] is Slot) && ((m_Slots[i] as Slot).Value is string))
                       {
                           string s = (m_Slots[i] as Slot).Value as string;
                           string[] words = s.Split(' ');

                           listOfAllWords.AddRange(words);
                       }


            return GetNextSymbol();
        }

        public override string GetNextSymbol()
        {
            if (current < listOfAllWords.Count)
               return listOfAllWords[current++];
            else
               return "";
        }

    }







    public abstract class CustomSlot : ConceptSystemElement
    {
        protected string m_Name;

        public string Name
        {
            get { return m_Name; }
        }

        public abstract string ToString(int indentation);
    }






    public class Slot: CustomSlot
    {
        private object m_Value;

        public Slot(string name, object value)
        {
            m_Name  = name;
            m_Value = value;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// The value of a slot can be anything.
        /// </summary>
        ///---------------------------------------------------------------------
        public object Value
        {
            get { return m_Value; }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Write a pair consisting of the name of the slot and its value
        /// </summary>
        /// <param name="indentation"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public override string ToString(int indentation)
        {
            if (m_Deleted)
               return string.Empty;

            string prefix = new string(' ', indentation);
            string result = ""; // Evtl. mit StringBuilder zu-tun

            result += prefix + m_Name + " " + ValueToString(m_Name.Length + 1 + indentation);

            return result;
        }

        private string ValueToString(int indentation)
        {
            string prefix = new string(' ', indentation);

            if (m_Value == null)
                return "NULL";
            else
            if (m_Value is string)
                return m_Value as string;
            else
            if (m_Value is Concept)
                return (m_Value as Concept).ToStringRestIndented(indentation);
            else
            if (m_Value is ConceptList)
                return (m_Value as ConceptList).ToStringRestIndented(indentation);
            else
            return "???";
        }

        public override string ToString()
        {
            return ToString(0);
        }
    }















    ///-------------------------------------------------------------------------
    /// <summary>
    /// Named list of concepts 
    /// </summary>
    ///-------------------------------------------------------------------------
    public class ConceptList : CustomSlot
    {
        private List<Concept> m_Concepts;

        public ConceptList(string name)
        {
            m_Name = name;
        }

        public List<Concept> Concepts
        {
            get { return m_Concepts; }
        }

        public void Add(Concept concept)
        {
            if (m_Concepts == null)
               m_Concepts = new List<Concept>();

            m_Concepts.Add(concept);
        }


        public string ToStringRestIndented(int indentation)
        {
            if (m_Deleted)
                return string.Empty;

            string prefix = new string(' ', indentation);
            string result = ""; // Evtl. mit StringBuilder zu-tun

            result += m_Name;


            if ((m_Concepts != null) && (m_Concepts.Count > 0))
            {
                result += System.Environment.NewLine;
                for (int i = 0; i < m_Concepts.Count; i++)
                    if (i < m_Concepts.Count - 1)
                        result += prefix + m_Concepts[i].ToString(indentation + 1) + System.Environment.NewLine;
                    else
                        result += prefix + m_Concepts[i].ToString(indentation + 1);
            }

            return result;
        }


        public override string ToString(int indentation)
        {
            if (m_Deleted)
                return string.Empty;

            string prefix = new string(' ', indentation);
            return prefix + ToStringRestIndented(indentation);
        }

        public override string ToString()
        {
            return ToString(0);
        }

        public override string GetFirstSymbol()
        {
            if (listOfAllWords == null)
                listOfAllWords = new List<string>(1 + m_Concepts.Count);
            else
                listOfAllWords.Clear();

            current = 0;

            if (m_SearchConceptNames)
                listOfAllWords.Add(m_Name);

            if (m_SearchSlotNames)
                for (int i = 0; i < m_Concepts.Count; i++)
                    listOfAllWords.Add(m_Concepts[i].Name);

            return GetNextSymbol();
        }

        public override string GetNextSymbol()
        {
            if (current < listOfAllWords.Count)
               return listOfAllWords[current++];
            else
               return "";
        }

    }


}
