//------------------------------------------------------------------------------
// Purpose: Inference engine for concepts
// Date   : 01.Sep.2007
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using Symps;
using Fusym;
using System.Collections;

namespace Symps
{
    internal class VariableBinding
    {
        private Hashtable m_Hashtable = new Hashtable();

        ///---------------------------------------------------------------------
        /// <summary>
        /// Add a new key / value pair. If the key already exists, nothing
        /// is done.
        /// </summary>
        ///---------------------------------------------------------------------
        public bool Add(string name, object value)
        {
            if (m_Hashtable.ContainsKey(name))
               return false;

           m_Hashtable.Add(name, value);
           return true;
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Return the binding of a variable. If that variable has not
        /// been bound before, return a pointer to the used instance
        /// of the class VariableBinding, in order to distinguish the
        /// result from a variable's value null.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public object Get(string name)
        {
            if (m_Hashtable.ContainsKey(name))
               return m_Hashtable[name];
            else
               return this;
        }
    }


    public class ConceptInferenceSystem
    {
        SemanticNetwork m_ClusterManager;

        public SemanticNetwork ClusterManager
        {
            set { m_ClusterManager = value; }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Tests if an object is a variable, that is a string with a preceding '?'.
        /// </summary>
        /// <param name="o"></param>
        /// <returns>The variable name with preceeding '?' as string, if it
        /// is a variable, otherwise null</returns>
        ///---------------------------------------------------------------------
        public static string IsVariable(object o)
        {
            string name;

            if (o is string)
                name = o as string;
            else
            if (o is Concept)
                name = (o as Concept).Name;
            else
                return null;

            if ((name.Length > 0) && (name[0] == '?'))
               return name;
            else
               return null;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Replace all variables in the consequence of the implication 
        /// with their value. The variable values must be in binding.
        /// </summary>
        /// <param name="consequence"></param>
        /// <param name="binding"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        internal Concept ApplyBinding(Concept consequence, VariableBinding binding)
        {
            Concept result = new Concept(consequence.Name);

            foreach (CustomSlot cs in consequence.Slots)
            {
                if (cs is ConceptList)
                   result.AddConceptList(cs as ConceptList);
                else
                {
                    Slot slot = cs as Slot;

                    string variable = IsVariable(slot.Value);
                    if (variable != null)
                    {
                        object value = binding.Get(variable);
                        if (value == binding)
                           return new Concept("UNBOUND-VARIABLE");

                       result.AddSlot(slot.Name, value);
                    }

                }
            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Entry to the recursive function Unify(3 parameters)
        /// </summary>
        ///---------------------------------------------------------------------
        internal VariableBinding Unify(Concept pattern, Concept instance)
        {
            VariableBinding result = new VariableBinding();
            return Unify(pattern, instance, result);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Determine under which conditions two concepts are equal. These conditions
        /// are returned as binding of variables.
        /// </summary>
        /// <param name="pattern"></param>
        /// <param name="instance"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        internal VariableBinding Unify(Concept pattern, Concept instance, VariableBinding result)
        {
            if (pattern.Name != instance.Name)
               return null;

            VariableBinding slotsOfInstance = GetSlotsOfConceptAsVariableBinding(instance);

            if (pattern.Slots == null)
               return result;

            //--- Search for slot values that are varibles in the concept pattern ---
            foreach (CustomSlot cs in pattern.Slots)
            {
                if (!(cs is Slot)) // A ConceptList doesn't play a role. In future versions maybe.
                    continue;

                Slot slot = cs as Slot;

                object valueOfSlotInInstance = slotsOfInstance.Get(slot.Name);

                //--------------------------------------------------------------
                // The special return value means slot.Name does not exist 
                // in instance. Therefore, we must reject the instance. 
                //--------------------------------------------------------------
                if (valueOfSlotInInstance == slotsOfInstance)
                   return null;


                string variable = IsVariable(slot.Value);
                if (variable != null)                            // That means it is a variable.
                   result.Add(variable, valueOfSlotInInstance);  // Bind it.
                else
                {
                    //----------------------------------------------------------
                    // If the value of the slot of pattern is a concept, the 
                    // corresponding value of the instance must also be 
                    // a concept. Recurse.
                    //----------------------------------------------------------
                    if ((slot.Value is Concept) && (valueOfSlotInInstance is Concept))
                        result = Unify(slot.Value as Concept, valueOfSlotInInstance as Concept, result);
                    else
                    //----------------------------------------------------------
                    // Any other type of value. Must be identical.
                    //----------------------------------------------------------
                    if (slot.Value != valueOfSlotInInstance)
                       return null;
                }
            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Get the pairs (Slot name, Slot value) for each slot of a concept
        /// as a VariableBinding object.
        /// </summary>
        /// <param name="c"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        private VariableBinding GetSlotsOfConceptAsVariableBinding(Concept c)
        {
            VariableBinding result = new VariableBinding();

            if (c.Slots == null)
               return result;

            foreach (CustomSlot cs in c.Slots)
            {
                if (! (cs is Slot)) // A ConceptList doesn't play a role.
                   continue;

                Slot slot = cs as Slot;

                result.Add(slot.Name, slot.Value);
            }

            return result;
        }






        









        ///---------------------------------------------------------------------
        /// <summary>
        /// Transforms a concept into another using a certain transformation rule.
        /// </summary>
        /// <param name="transformation">A pair of concepts, the old and the new pattern</param>
        /// <param name="concept">The starting concept to be transformed</param>
        ///---------------------------------------------------------------------
        public Concept Implication(ConceptList transformation, Concept concept)
        {
            List<Concept> list = transformation.Concepts;

            if (list.Count != 2)
               return new Concept("TWO-PARAMETERS-EXPECTED");

            Concept condition   = list[0];
            Concept consequence = list[1];

            VariableBinding binding = Unify(condition, concept);

            if (binding != null)
               return ApplyBinding(consequence, binding);
            else
               return new Concept("CONCEPT-TRANSFORMATION-NOT-APPLICABLE");
        }



        public Concept Transform(ConceptList possibleTransformation, Concept concept)
        {
            string name = possibleTransformation.Name;

            if ((name == "->") || (name == "IF-THEN"))
               return Implication(possibleTransformation, concept);
            else
               return new Concept("NOT-A-KNOWN-CONCEPT-TRANSFORMATION");
        }

    }
}
