//------------------------------------------------------------------------------
// Purpose: Assigns a value to each document.
// Date   : 26.Apr.2006
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Text.RegularExpressions;

namespace Fusym
{






    ///-------------------------------------------------------------------------
    /// <summary>
    /// Creates an associative ring around a list of symbols.
    /// </summary>
    ///-------------------------------------------------------------------------
    public class DocumentEvaluator : IDocumentEvaluator
    {
        SemanticNetwork   m_SemanticNetwork;
        public DistanceManager  DistanceManager;
        SemanticDistanceByEdgeNumber m_SemanticDistanceByEdgeNumber;
        List<string>     m_inputSymbols;
        int              m_depth;
        int              m_maxCount;
        int              m_CountOfInputSymbols;
        List<string>     m_AssociatedSymbols = new List<string>();
        Dictionary<string,EvaluationVector> m_SymbolsToVectors = new Dictionary<string,EvaluationVector>();
        EvaluationMethod m_EvaluationMethod;
        List<int>        m_DistanceTable;
        int              m_DistanceTableEntryMax;
        bool             m_FullTextSearch;
        private string   m_Symbols;

        private delegate double EvaluationFunction(EvaluationVector vector);
        private EvaluationFunction m_EvaluationFunction;

        public string Symbols
        {
            get { return m_Symbols; }
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Constructor.
        /// </summary>
        ///---------------------------------------------------------------------
        public DocumentEvaluator(
                                 SemanticNetwork semanticNetwork, 
                                 List<string>    inputSymbols, 
                                 int depth, 
                                 int maxCount,
                                 EvaluationMethod evaluationMethod, 
                                 bool fullTextSearch
                                )
        {
            m_SemanticNetwork      = semanticNetwork;
            m_SemanticDistanceByEdgeNumber = new SemanticDistanceByEdgeNumber(m_SemanticNetwork);
            m_inputSymbols        = inputSymbols;
            m_CountOfInputSymbols = inputSymbols.Count;
            m_depth               = depth;
            m_maxCount            = maxCount;
            m_EvaluationMethod    = evaluationMethod;
            m_FullTextSearch      = fullTextSearch;



            GetDistances(m_AssociatedSymbols);

            if (
                (m_EvaluationMethod == EvaluationMethod.Fusym) ||
                (m_EvaluationMethod == EvaluationMethod.EdgeSum) ||
                (m_EvaluationMethod == EvaluationMethod.CloudMatching)
               )
               {
                    InitDistanceTable();
                    InitDistances();
               }



            // Initialize the function pointer m_EvaluationFunction
           if (m_EvaluationMethod == EvaluationMethod.Sum)
               m_EvaluationFunction = EvaluateSum;
           else
           if (m_EvaluationMethod == EvaluationMethod.SquareSum)
               m_EvaluationFunction = EvaluateSquareSum;
           else
           if (m_EvaluationMethod == EvaluationMethod.Fusym)
               m_EvaluationFunction = EvaluateFusym;
           else
           if (m_EvaluationMethod == EvaluationMethod.EdgeSum)
               m_EvaluationFunction = EvaluateFusym;
           else
           if (m_EvaluationMethod == EvaluationMethod.CloudMatching)
              m_EvaluationFunction = EvaluateFusym;
           else
               throw new Exception("DocumentEvaluator: Unexpected evaluation method");
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Use this list to retrieve documents containing each of the strings.
        /// </summary>
        /// <returns>A stringlist containing strings that are
        /// associated with the input symbols</returns>
        ///---------------------------------------------------------------------
        public List<string> AssociatedSymbols
        {
            get { 
                    return m_AssociatedSymbols; 
                }
        } 


        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculate the value of a document. Sets the member variables
        /// EvaluationResult and Symbols.
        /// </summary>
        ///---------------------------------------------------------------------
        public double Evaluate(IDocument document)
        {
            EvaluationVector vector = new EvaluationVector(m_inputSymbols.Count, m_depth);

            DistanceVectorForDocument(document, vector);
            m_Symbols = vector.ToString(); // Wohl Info, warum Dokument so bewertet wurde.
            return m_EvaluationFunction(vector); // Indirect function call
        }


        public double EvaluateSum(EvaluationVector vector)
        {
            double result = 0.0;

            for (int i = 0; i < vector.Count; i++)
                result += vector.DistanceList[i];

            return result;
        }


        public double EvaluateSquareSum(EvaluationVector vector)
        {
            double result = 0.0;

            for (int i = 0; i < vector.Count; i++)
            {
                int component = vector.DistanceList[i];
                result += component * component;
            }

            return result;
        }

        /*
        
        public double EvalueFusym(List<int> vector)
        {
            double result = 0.0;
            int index = 0;

            if (m_DistanceTableEntryMax <= 1)
               return EvaluateSum(vector);


            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = 0; j < m_CountOfInputSymbols; j++)
                    if (i < j)
                    {
                        int theDistance = m_DistanceTable[index];

                        int value1 = vector[i];
                        int value2 = vector[j];

            //            double ratio = theDistance / (m_DistanceTableEntryMax); // Hier kann man variieren zu-tun
                        double ratio = theDistance / (m_DistanceTableEntryMax + 1); // Hier kann man variieren zu-tun
                        int min = minI(value1, value2);
                        int max = maxI(value1, value2);

                        result += max * (1.0 - ratio) + min * ratio;

                        index++;
                    }

            return result;
        }
        


        */





        public double EvaluateFusym(EvaluationVector vector)
        {
            double result = 0.0;
            int index = 0; // Indicates current pair of input symbols.

            if (m_DistanceTableEntryMax <= 1)
                return EvaluateSum(vector);

            //alt:            double inverse = 1 / (double)(m_DistanceTableEntryMax + 1);
            double inverse = 1 / (double)m_DistanceTableEntryMax;

            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = i + 1; j < m_CountOfInputSymbols; j++)
                {
                    double theDistance = m_DistanceTable[index];

                    int value1 = vector.DistanceList[i];
                    int value2 = vector.DistanceList[j];

                    //            double ratio = theDistance / (m_DistanceTableEntryMax); // Hier kann man variieren zu-tun
                    double ratio = theDistance * inverse;
                    double min = Math.Min(value1, value2);
                    double max = Math.Max(value1, value2);

                    // alt                  result += max * (1.0 - ratio) + min * ratio;
                    result += min * (1.0 - ratio) + max * ratio;

                    index++;
                }

            return result;
        }

        
        private double EvalueEdgeSum(EvaluationVector vector)
        {
            throw new Exception("The method or operation is not implemented.");
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Retrieve the semantic distance between two symbols.
        /// </summary>
        /// <param name="symbol1">A symbol</param>
        /// <param name="symbol2">A symbol</param>
        /// <returns>The semantic distance, or -1 in case of no found path</returns>
        ///---------------------------------------------------------------------
        private int SemanticDistance(string symbol1, string symbol2)
        {
            if (
                (m_EvaluationMethod == EvaluationMethod.Sum      ) ||
                (m_EvaluationMethod == EvaluationMethod.SquareSum) ||
                (m_EvaluationMethod == EvaluationMethod.Fusym    ) 
               )
               return m_SemanticNetwork.SemanticDistance(symbol1, symbol2);
            else
            if (m_EvaluationMethod == EvaluationMethod.EdgeSum)
               return m_SemanticDistanceByEdgeNumber.SemanticDistance(symbol1, symbol2);
            else
            if (m_EvaluationMethod == EvaluationMethod.CloudMatching)
               return m_SemanticNetwork.SemanticDistanceByCloudMatching(symbol1, symbol2, m_maxCount, m_depth);
            else
               throw new Exception("Unexpected evaluation method");
            
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Init distance matrix. As a side effect, the maximum distance
        /// is determined.
        /// </summary>
        ///---------------------------------------------------------------------
        private void InitDistanceTable()
        {
            m_DistanceTable = new List<int>(m_CountOfInputSymbols * m_CountOfInputSymbols);
            m_DistanceTableEntryMax = 0;
            int semanticDistance;

            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = i + 1; j < m_CountOfInputSymbols; j++)
                {
                    semanticDistance = SemanticDistance(m_inputSymbols[i], m_inputSymbols[j]);
                    if (semanticDistance < 0)      // -1 means: no path found
                        semanticDistance = m_depth; // therefore assume the maximum.


                    m_DistanceTable.Add(semanticDistance);

                    if (semanticDistance > m_DistanceTableEntryMax)
                        m_DistanceTableEntryMax = semanticDistance;
                }
        }
        
        
        ///---------------------------------------------------------------------
        /// <summary>
        /// Init the distances table. 
        /// </summary>
        ///---------------------------------------------------------------------
        private void InitDistances()
        {
            int index = 0;
            
           DistanceManager = new DistanceManager(m_DistanceTableEntryMax);
        
            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = 0; j < m_CountOfInputSymbols; j++)
                    if (i < j)
                    {
                        int theDistance = m_DistanceTable[index];
                        DistanceManager.Add(m_inputSymbols[i],m_inputSymbols[j],theDistance);                        
                        index++;
                    }
        }
        
        
        
        
        

        /*

        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculates an associative ring around one symbol.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> Circle(string symbol)
        {
            List<Node> list = new List<Node>();

            int count = 0;
            int depth = 0;

            list.Clear();

            Node n = m_clusterManager.GetNode(symbol);
            if (n == null)
               return list;
            

            Queue<Node> queue = new Queue<Node>();
            
            m_clusterManager.MarkAllNodes(-1);
            n.Mark = 0;
            queue.Enqueue(n);

            while (queue.Count != 0)
            {
                n = queue.Dequeue();

                depth = n.Mark;
                count++;

                if ((count <= m_maxCount) && (depth++ <= m_depth))
                    list.Add(n);
                else
                    return list;

                // Iteration through all m_Neighbours of the current node n
                foreach (Node m in n.Neighbours)
                {
                    if (m.Mark < 0) // Not already marked?
                       {
                           m.Mark = depth;
                           queue.Enqueue(m);
                       }
                }         
            }

            return list;
        }
        
        
        */
        
        
        
        
        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculates an associative ring around one symbol.
        /// </summary>
        ///---------------------------------------------------------------------
        /*
        public void GetCircleRandomWalk(List<string> symbols, List<Node> list)
        {
            Random rndObj = new Random();
            
            list.Clear();

              
            m_SemanticNetwork.MarkAllNodes(0);


            foreach (string symbol in symbols)
            {
                Node startNode = m_SemanticNetwork.GetNode(symbol);

                int count = 0;
                while (count <= m_maxCount)
                {
                    Node n = startNode;
                    int depth = 0;

                    while (depth <= m_depth)
                    {
                        if (n.Mark == 0)
                            list.Add(n);

                        n.Mark++;
                        depth++;
                        count++;

                        // Choose a random neighbour
                        int neighborcount = n.Neighbours.Count;
                        if (neighborcount < 1)
                            continue;

                        int idx = rndObj.Next(neighborcount);
                        n = n.Neighbours[idx];
                    }
                }

            }


            // A kind of normalization             
            int maxMark = MaxMark(list);
            foreach (Node m in list)
            {
                m.Mark = maxMark - m.Mark;
            }       
        }       
                  */
                  
                  
        ///---------------------------------------------------------------------
        /// <summary>
        /// Determine the maximal Mark value
        /// </summary>
        ///---------------------------------------------------------------------
        public int MaxMark(List<Node> list)
        {
            if (list.Count == 0)
               return 0;
               
            int result = list[0].Mark; // It is guaranteed that there are at least one elements
            
            for (int i = 1; i < list.Count; i++)
                if (list[i].Mark > result)
                   result = list[i].Mark;
            
            return result;
        }  
        
        


        ///---------------------------------------------------------------------
        /// <summary>
        /// Get a vector that has the same count of members as the input symbols.
        /// </summary>
        /// <returns>An ArrayList with the components == m_depth</returns>
        ///---------------------------------------------------------------------
        /*
        private List<int> GetNewVector()
        {
            List<int> result = new List<int>(m_inputSymbols.Count);
            for (int i = 0; i < m_inputSymbols.Count; i++) // Koennte man wegoptimieren zu-tun
                result.Add(m_depth);

            return result;
        }*/



        ///---------------------------------------------------------------------
        /// <summary>
        /// Retrieves a list of symbols that are associated with the input
        /// symbols given in the constructor. Also, a hashtable is
        /// calculated that maps those symbols to their semantic
        /// distances to each input symbols.
        /// </summary>
        /// <param name="extendedSymbols">A return parameter that must be
        /// instantiated by the caller. Other symbols that are associated with the 
        /// symbols given in the constructor. You can use the list for a search in an index.</param>
        ///---------------------------------------------------------------------
        public void GetDistances(List<string> extendedSymbols)
        {
            int countOfInputSymbols = m_inputSymbols.Count;
            List<Node> list;
            int depth = m_depth;




            for (int i = 0; i < m_inputSymbols.Count; i++)
            {
                string s = m_inputSymbols[i];

                // Collect neighbour symbols with their Mark set.
                if (m_EvaluationMethod != EvaluationMethod.EdgeSum)
                {
                    //list = m_SemanticNetwork.Circle(s, m_maxCount, depth);

                    
                    RandomWalk rw = new RandomWalk(m_SemanticNetwork);
                    //list = rw.GetCircleRandomWalkBeta(s, m_maxCount, depth);
                    list = rw.GetCircle_RW_SelfAvoiding(m_SemanticNetwork.AddSymbol(s), m_maxCount, depth, false);
                     


                    // A correction to m_depth for the case the depth
                    // is limited by the allowed count of symbols.
                    // This applies only to evaluation methods != EdgeSum,
                    // because at EdgeSum, the Mark has a different meaning.
                    int maxMark = MaxMark(list);
                    if (maxMark > depth)
                        depth = maxMark;

                }
                else
                    list = m_SemanticDistanceByEdgeNumber.Circle(s, depth); //zunchst verworfen, da zu kompliziert
                //                   m_SemanticDistanceByEdgeNumber.GetCircle(s,MaxMark(list),list); //zunchst verworfen, da zu kompliziert



                foreach (Node n in list)
                {
                    string sym = n.ToString(); // The string representation of the symbol

                    EvaluationVector vector;

                    /*
                    //--- Get the vector of sym, if already known ---
                    if (m_SymbolsToVectors.ContainsKey(sym))
                        vector = m_SymbolsToVectors[sym];
                    else
                    //--- A new symbol ---
                    {
                        vector = new EvaluationVector(m_inputSymbols.Count, depth);
                        m_SymbolsToVectors[sym] = vector;
                        extendedSymbols.Add(sym);
                    }*/


                    //--- Get the vector of sym, if already known ---
                    if (m_SymbolsToVectors.TryGetValue(sym, out vector) == false)
                    {
                        vector = new EvaluationVector(m_inputSymbols.Count, depth);
                        m_SymbolsToVectors[sym] = vector;
                        extendedSymbols.Add(sym);
                    }

                    vector.DistanceList[i] = n.Mark;
                    vector.Symbols[i] = n.ToString();
                }
            }
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Enumerates all symbols in the document. Determines the minimum
        /// distance to each query symbol that was provided in the constructor.
        /// </summary>
        /// <param name="theDocument">A document whose symbols are enumeratable</param>
        /// <param name="result">A distance vector</param>
        ///---------------------------------------------------------------------
        public void DistanceVectorForDocument(IDocument theDocument, EvaluationVector result)
        {
            string current = theDocument.GetFirstSymbol();

            while (current != "")
            {
                current = current.ToUpper();

            
                //--- Has current an own vector object? ---
                
                EvaluationVector theVector;

                if (m_SymbolsToVectors.TryGetValue(current, out theVector) == false)
                {
                    // Another chance: fulltext search might apply.
                    if (m_FullTextSearch == true)
                    {
                        theVector = null;
                        for (int i = 0; i < m_CountOfInputSymbols; i++)
                        {
                            string s = m_inputSymbols[i];
                            if (current.Contains(s))
                            {
                                if (m_SymbolsToVectors.TryGetValue(s, out theVector))
                                    break;
                            }
                        }
                    }
                }

                //--- Is the symbol in the range of the circle? ---
                if (theVector != null)
                   result.MergeMinima(theVector);
                
                current = theDocument.GetNextSymbol();
            }
        }
              
    }


}
