//------------------------------------------------------------------------------
// Purpose: Managing associations in Fusym technology.
// Date   : 26.Apr.2006
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;

namespace Fusym
{
    public enum EvaluationMethod
    {
        Sum       = 0,
        SquareSum = 1,
        Fusym     = 2
    }

    public interface IClusterProvider
    {
        List<string> GetFirstCluster();
        List<string> GetNextCluster();
    }






    ///-------------------------------------------------------------------------
    /// <summary>
    /// A node in the associative graph represents one symbol.
    /// </summary>
    ///-------------------------------------------------------------------------
    public class Node
    {
        string m_symbol;
        List<Node> neighbours = new List<Node>();
        Hashtable neighbourHash = new Hashtable();
        int CurrentNeighbour; // For enumeration
        int m_Mark;
        
        public List<Node> Neighbours
        {
            get { return neighbours; }
        }
        

        public Node(string symbol)
        {
            m_symbol = symbol;
        }


        public int Mark
        {
         get {return m_Mark;}
         set {m_Mark = value;}
        }

        public void AddNeighbour(Node n)
        {
            if (!neighbourHash.Contains(n))
               {
                   neighbours.Add(n);
                   neighbourHash.Add(n,1);
               }
        }

        public override string ToString()
        {
            return m_symbol;
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Enumeration of neighbour symbols.
        /// </summary>
        /// <returns>The first neighour, if there is one, null otherwise</returns>
        ///---------------------------------------------------------------------
        public Node GetFirstNeighbour()
        {
            CurrentNeighbour = 0;
            return GetNextNeighbour();
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Enumeration of neighbour symbols.
        /// </summary>
        /// <returns>The next neighour, if there is one, null otherwise</returns>
        ///---------------------------------------------------------------------
        public Node GetNextNeighbour()
        {
            if (CurrentNeighbour < neighbours.Count)
                return (Node)neighbours[CurrentNeighbour++];
            else
                return null;
        }
    }






















    ///-------------------------------------------------------------------------
    /// <summary>
    /// Creates an associative ring around a list of symbols.
    /// </summary>
    ///-------------------------------------------------------------------------
    public class DocumentEvaluator
    {
        ClusterManager   m_clusterManager;
        List<string>     m_inputSymbols;
        int              m_depth;
        int              m_maxCount;
        int              m_CountOfInputSymbols;
        List<string>     m_AssociatedSymbols = new List<string>(); 
        Hashtable        m_SymbolsToVectors = new Hashtable();
        EvaluationMethod m_EvaluationMethod;
        List<int>        m_DistanceTable;
        int              m_DistanceTableEntryMax;


        public DocumentEvaluator(ClusterManager clusterManager, List<string> inputSymbols, int depth, int maxCount, EvaluationMethod evaluationMethod)
        {
            m_clusterManager      = clusterManager;
            m_inputSymbols        = inputSymbols;  // Eindeutig machen zu-tun
            m_CountOfInputSymbols = inputSymbols.Count;
            m_depth               = depth;
            m_maxCount            = maxCount;
            m_EvaluationMethod    = evaluationMethod;

            GetDistances(m_AssociatedSymbols);


            if (m_EvaluationMethod == EvaluationMethod.Fusym)
               InitDistanceTable();
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Use this list to retrieve documents containing each of the strings.
        /// </summary>
        /// <returns>A stringlist containing strings that are
        /// associated with the input symbols</returns>
        ///---------------------------------------------------------------------
        public List<string> GetAssociatedSymbols()
        {
            return m_AssociatedSymbols;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculate the value of a document
        /// </summary>
        /// <param name="document"></param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public double Evaluate(IDocument document, out string symbols)
        {
            EvaluationVector vector = new EvaluationVector(m_inputSymbols.Count, m_depth);

            GetDistanceVectorForADocument(document, vector);
            symbols = vector.ToString();

            if (m_EvaluationMethod == EvaluationMethod.Sum)
                return EvaluateSum(vector);
            else
            if (m_EvaluationMethod == EvaluationMethod.SquareSum)
               return EvaluateSquareSum(vector);
            else
            if (m_EvaluationMethod == EvaluationMethod.Fusym)
               return EvalueFusym(vector);
            else
               throw new Exception("Unexpected evaluation method");
        }


        public double Evaluate(IDocument document)
        {
            string s;
            return Evaluate(document, out s);
        }



        public double EvaluateSum(EvaluationVector vector)
        {
            double result = 0.0;

            for (int i = 0; i < vector.Count; i++)
                result = result + vector.GetIntComponent(i);

            return result;
        }


        public double EvaluateSquareSum(EvaluationVector vector)
        {
            double result = 0.0;

            for (int i = 0; i < vector.Count; i++)
            {
                int component = vector.GetIntComponent(i);
                result = result + component * component;
            }

            return result;
        }

        /*
        
        public double EvalueFusym(List<int> vector)
        {
            double result = 0.0;
            int index = 0;

            if (m_DistanceTableEntryMax <= 1)
               return EvaluateSum(vector);


            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = 0; j < m_CountOfInputSymbols; j++)
                    if (i < j)
                    {
                        int theDistance = m_DistanceTable[index];

                        int value1 = vector[i];
                        int value2 = vector[j];

            //            double ratio = theDistance / (m_DistanceTableEntryMax); // Hier kann man variieren zu-tun
                        double ratio = theDistance / (m_DistanceTableEntryMax + 1); // Hier kann man variieren zu-tun
                        int min = minI(value1, value2);
                        int max = maxI(value1, value2);

                        result += max * (1.0 - ratio) + min * ratio;

                        index++;
                    }

            return result;
        }
        


        */





        public double EvalueFusym(EvaluationVector vector)
        {
            double result = 0.0;
            int index = 0;

            if (m_DistanceTableEntryMax <= 1)
                return EvaluateSum(vector);


            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = 0; j < m_CountOfInputSymbols; j++)
                    if (i < j)
                    {
                        double theDistance = m_DistanceTable[index];

                        int value1 = vector.GetIntComponent(i);
                        int value2 = vector.GetIntComponent(j);

                        //            double ratio = theDistance / (m_DistanceTableEntryMax); // Hier kann man variieren zu-tun
                        double ratio = theDistance / (m_DistanceTableEntryMax + 1); // Hier kann man variieren zu-tun
                        double min = minI(value1, value2);
                        double max = maxI(value1, value2);

                        result += max * (1.0 - ratio) + min * ratio;

                        index++;
                    }

            return result;
        }




        






        ///---------------------------------------------------------------------
        /// <summary>
        /// Retrieve the semantic distance.
        /// </summary>
        /// <param name="symbol1">A symbol</param>
        /// <param name="symbol2">A symbol</param>
        /// <returns>The semantic distance, or -1 in case of no found path</returns>
        ///---------------------------------------------------------------------
        private int GetSemanticDistance(string symbol1, string symbol2)
        {
            return m_clusterManager.SemanticDistance(symbol1, symbol2);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Init distance matrix
        /// </summary>
        ///---------------------------------------------------------------------
        private void InitDistanceTable()
        {
            m_DistanceTable = new List<int>(m_CountOfInputSymbols * m_CountOfInputSymbols);
            m_DistanceTableEntryMax = 0;

            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = 0; j < m_CountOfInputSymbols; j++)
                    if (i < j)
                       {
                           int theDistance = GetSemanticDistance(m_inputSymbols[i], m_inputSymbols[j]);
                           m_DistanceTable.Add(theDistance);

                           if (theDistance > m_DistanceTableEntryMax)
                              m_DistanceTableEntryMax = theDistance;
                       }
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculates an associative ring around one symbol.
        /// </summary>
        /// <param name="symbol">The symbol as string</param>
        /// <param name="list">A list of symbol nodes with their Mark property set to the distance to symbol</param>
        ///---------------------------------------------------------------------
        public void GetCircle(string symbol, List<Node> list)
        {
            int count = 0;
            int depth = 0;

            list.Clear();

            Node n = m_clusterManager.GetNode(symbol);
            if (n == null)
               return;
            

            Queue<Node> queue = new Queue<Node>();

            m_clusterManager.MarkAllNodes(-1);
            n.Mark = 0;
            queue.Enqueue(n);

            while (queue.Count != 0)
            {
                n = queue.Dequeue();

                depth = n.Mark;
                count++;

                if ((count <= m_maxCount) && (depth++ <= m_depth))
                    list.Add(n);
                else
                    return;

                Node m = n.GetFirstNeighbour();
                while (m != null)
                {
                    if (m.Mark < 0) // Not already marked?
                       {
                           m.Mark = depth;
                           queue.Enqueue(m);
                       }
                    m = n.GetNextNeighbour();
                }
            }

        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Get a vector that has the same count of members as the input symbols.
        /// </summary>
        /// <returns>An ArrayList with the components == m_depth</returns>
        ///---------------------------------------------------------------------
        /*
        private List<int> GetNewVector()
        {
            List<int> result = new List<int>(m_inputSymbols.Count);
            for (int i = 0; i < m_inputSymbols.Count; i++) // Koennte man wegoptimieren zu-tun
                result.Add(m_depth);

            return result;
        }*/



        ///---------------------------------------------------------------------
        /// <summary>
        /// Retrieves a list of symbols that are associated with the input
        /// symbols given in the constructor. Also, a hashtable is
        /// calculated that maps those symbols to their semantic
        /// distances to each input symbols.
        /// </summary>
        /// <param name="extendedSymbols">A return parameter that must be
        /// instantiated by the caller. Other symbols that are associated with the 
        /// symbols given in the constructor. You can use the list for a search in an index.</param>
        ///---------------------------------------------------------------------
        public void GetDistances(List<string> extendedSymbols)
        {
            int countOfInputSymbols = m_inputSymbols.Count;
            List<Node> list = new List<Node>();

            for (int i = 0; i < m_inputSymbols.Count; i++)
            {
                string s = m_inputSymbols[i];

                list.Clear();
                GetCircle(s, list); // Collect neighbour symbols with their Mark set.
                foreach (Node n in list)
                {
                    string sym = n.ToString(); // The string reprentation of the symbol

                    EvaluationVector vector;

                    //--- Get the vector of sym, if already known ---
                    if (m_SymbolsToVectors.Contains(sym))
                        vector = (EvaluationVector)m_SymbolsToVectors[sym];
                    else
                    //--- A new symbol ---
                       {
                           vector = new EvaluationVector(m_inputSymbols.Count, m_depth);
                           m_SymbolsToVectors[sym] = vector;
                           extendedSymbols.Add(sym);
                       }

                    vector.SetIntComponent(i,n.Mark);
                    vector.SetStringComponent(i, n.ToString());
                }
            }
        }


        private int minI(int a, int b)
        {
            if (a < b)
                return a;
            else
                return b;
        }


        private int maxI(int a, int b)
        {
            if (a > b)
                return a;
            else
                return b;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Enumerates all symbols in the document. Determines the minimum
        /// distance to each query symbol that was provided in the constructor.
        /// </summary>
        /// <param name="theDocument">A document whose symbols are enumeratable</param>
        /// <param name="result">A distance vector</param>
        ///---------------------------------------------------------------------
        public void GetDistanceVectorForADocument(IDocument theDocument, EvaluationVector result)
        {
            string current;

            current = theDocument.GetFirstSymbol();

            while (current != "")
            {
                //--- Has current an own vector object? ---
                EvaluationVector theVector = (EvaluationVector)m_SymbolsToVectors[current.ToUpper()];

                //--- Is the symbol in the range of the circle? ---
                if (theVector != null)
                   result.MergeMinima(theVector);
                
                current = theDocument.GetNextSymbol();
            }
        }
              
    }























    ///-------------------------------------------------------------------------
    /// <summary>
    /// Represents the associative network
    /// </summary>
    ///-------------------------------------------------------------------------
    public class ClusterManager : IDisposable
    {
        Hashtable symbolHash  = new Hashtable();
        List<Node> m_NodeList = new List<Node>();
        Encoding  m_Encoding;


        public void Dispose()
        {
            return;
        }


        public ClusterManager(Encoding encoding)
        {
            m_Encoding = encoding;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Add an edge between two Nodes in both directions.
        /// </summary>
        /// <param name="sym1">A symbol</param>
        /// <param name="sym2">A symbol</param>
        ///---------------------------------------------------------------------
        protected void AddPair(Node sym1, Node sym2)
        {
            sym1.AddNeighbour(sym2);
            sym2.AddNeighbour(sym1);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Add a list of nodes to the graph.
        /// Each pair (order is irrelevant) builds up an edge.
        /// </summary>
        /// <param name="cluster">A list of symbols (strings).</param>
        ///---------------------------------------------------------------------
        protected void AddCluster(List<Node> cluster)
        {
            if (cluster.Count <= 0)
               return;

          for (int i = 0; i < cluster.Count-1; i++)
              for (int j = i+1; j < cluster.Count; j++)
                  if (j > i)
                     AddPair((Node)cluster[i], (Node)cluster[j]);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Makes a symbol known.
        /// </summary>
        /// <param name="symbol">An uppercase string without spaces, a symbol</param>
        /// <returns>The Node object of the symbol.</returns>
        ///---------------------------------------------------------------------
        protected Node AddSymbol(string symbol)
        {
            Node result;

            if (symbolHash.Contains(symbol))
                return (Node)symbolHash[symbol];

            result = new Node(symbol);
            symbolHash[symbol] = result;
            m_NodeList.Add(result);
            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Gives a Node instance for a symbol string.
        /// </summary>
        /// <param name="symbol">String representation of the symbol</param>
        /// <returns>A Node instance, if there is one, null otherwise</returns>
        ///---------------------------------------------------------------------
        public Node GetNode(string symbol)
        {
            return (Node)symbolHash[symbol.ToUpper()];
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Check if a symbol is known and has at least one neighbour
        /// </summary>
        /// <param name="symbol">String representation of the symbol</param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public bool HasAssociations(string symbol)
        {
            Node n = (Node)symbolHash[symbol.ToUpper()];

            if (n == null)
                return false;
            else
                return n.GetFirstNeighbour() != null;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Loads a cluster file and integrates its associations into the graph
        /// </summary>
        /// <param name="filename">A filename</param>
        ///---------------------------------------------------------------------
        public void LoadFromFile(string filename)
        {
            List<Node> cluster  = new  List<Node>();
            Regex regexSpace   = new Regex(@"^\s*$");
            Regex regexNoSpace = new Regex(@"^\s*\S+\s*$");

            StreamReader sr = null;
            try
            {
             sr = new StreamReader(new BufferedStream(new FileStream(filename, FileMode.Open)), m_Encoding);
            							
             string line = "";


             while((line = sr.ReadLine()) != null)
             {
                 if (regexNoSpace.IsMatch(line))
                    {
                        Node sym;
                        sym = AddSymbol(line.ToUpper().Trim());
                        cluster.Add(sym);
                    }
                 else
                 if (regexSpace.IsMatch(line))
                    {
                     AddCluster(cluster);
                     cluster.Clear();
                    }
                 /* 
                 else
                   unknown line type.
                 */
             }

             AddCluster(cluster);
             cluster.Clear();
         }
         finally
         {
          sr.Close();
         }
        }



        public void LoadFromClusterProvider(IClusterProvider clusterProvider)
        {
            if (clusterProvider == null)
               return;

            List<string> stringList = clusterProvider.GetFirstCluster();
            List<Node> cluster    = new List<Node>();
            while (stringList != null)
            {
                cluster.Clear();
                for (int i = 0; i < stringList.Count; i++)
                {
                    Node sym = AddSymbol(stringList[i]);
                    cluster.Add(sym);
                }


                AddCluster(cluster);
                stringList = clusterProvider.GetNextCluster();
            }
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Set the Mark property of all Node objects to value.
        /// </summary>
        /// <param name="value">The int value</param>
        ///---------------------------------------------------------------------
        public void MarkAllNodes(int value)
        {
            foreach (Node n in m_NodeList)
            {
                n.Mark = value;
            }
        }




        ///---------------------------------------------------------------------
        /// <summary>
        /// Find a path from one symbol to another in the graph.
        /// </summary>
        /// <param name="start">A symbol string</param>
        /// <param name="end">Another symbol string</param>
        /// <param name="lis">List that will be filled with a list of Node objects on the path</param>
        /// <returns>true if a path could be found, false otherwise</returns>
        ///---------------------------------------------------------------------
        public bool FindPath(string start, string end, List<Node> lis)
        {
            Node startNode = GetNode(start);
            Node endNode   = GetNode(end);
            Node current;
            Queue q        = new Queue();
            bool found     = false;
            int nextMark;
            lis.Clear();

            if ((startNode == null) || (endNode == null))
                return false;

            found = false;
            MarkAllNodes(-1);
            startNode.Mark = 0;
            q.Enqueue(startNode);
            while (q.Count != 0)
            {
                current = (Node)q.Dequeue();
                if (current == endNode)
                {
                    found = true;
                    break;
                }

                // Mark all neighbours of current
                nextMark = current.Mark + 1;
                Node m = current.GetFirstNeighbour();
                while (m != null)
                {
                    if (m.Mark < 0) // Not already visited
                       {
                        m.Mark = nextMark;
                        q.Enqueue(m);
                       }
                    m = current.GetNextNeighbour();
                }
            }



            if (!found)
               return false;



            // Now step back from endNode to startNode
            current = endNode; // To make the compiler happy
            lis.Add(endNode);
            nextMark = endNode.Mark - 1;
            while (current != startNode)
            {
                // Search for a neighbour of currentnode with Mark == nextMark
                Node m = current.GetFirstNeighbour();
                while (m != null)
                {
                    if (m.Mark == nextMark) // Not already visited
                       {
                        current = m;
                        lis.Add(current);
                        nextMark--;
                        break;
                       }
                    m = current.GetNextNeighbour();
                }
            }

            lis.Reverse();
            return true;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculate the semantic distance between two symbols.
        /// 
        /// Man kann hier optimieren; der Rueckweg ist uninteressant.
        /// </summary>
        /// <param name="symbol1">One symbol</param>
        /// <param name="symbol2">Another symbol</param>
        /// <returns>The length of the path in the graph between the symbols, -1 if there is no path</returns>
        ///---------------------------------------------------------------------
        public int SemanticDistance(string symbol1, string symbol2)
        {
            List<Node> lis = new List<Node>();

            if (FindPath(symbol1, symbol2, lis))
               return lis.Count - 1;
            else
               return -1;
        }


        public DocumentEvaluator GetDocumentEvaluator(List<string> inputSymbols, int depth, int maxCount, EvaluationMethod evaluationMethod)
        {
            return new DocumentEvaluator(this, inputSymbols, depth, maxCount, evaluationMethod);
        }

        
    }
}
