//------------------------------------------------------------------------------
// Purpose: Random walk on a semantic network.
// Date   : 09. Jul. 2011
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace Fusym
{
    public class RandomWalk
    {
        private SemanticNetwork m_SemanticNetwork;
        private Random rndObj;
        private int m_CountOfInputSymbols;
        private List<Dictionary<Node, double>> m_DictionariesForRanks;
        private List<Dictionary<Node, int>> m_DictionariesForRanks2;
        private List<List<NodeAndMark>> m_Alpha;
        private double[] m_PairDistances_0_1;

        public double[] PairDistances_0_1
        {
            get { return m_PairDistances_0_1; }
            set { m_PairDistances_0_1 = value; }
        }
        private Dictionary<string, EvaluationVector> m_DictWordToEvaluationVector;

        public Dictionary<string, EvaluationVector> DictWordToEvaluationVector
        {
            get { return m_DictWordToEvaluationVector; }
            set { m_DictWordToEvaluationVector = value; }
        }
        private List<Node> inputSymbols;

        ///---------------------------------------------------------------------
        /// <summary>
        /// Constructor.
        /// </summary>
        ///---------------------------------------------------------------------
        public RandomWalk(SemanticNetwork semanticNetwork)
        {
            rndObj = new Random();
            m_SemanticNetwork = semanticNetwork;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Convert a stringlist into a node list.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> StringsToNodes(List<string> symbols)
        {
            List<Node> result = new List<Node>();
            foreach (string s in symbols)
            {
                Node n = m_SemanticNetwork.GetNode(s);
                if (n == null)
                    n = m_SemanticNetwork.AddSymbol("???");

                result.Add(n);
            }
            return result;
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Takes a list of input symbols and returns a list of nodes with 
        /// mark value set.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> GetCircleRandomWalk(List<string> symbols, int maxCount, int maxDepth, bool normalize)
        {
            List<Node> nodes = StringsToNodes(symbols);

            List<Node> result = RW_SelfAvoiding(nodes, maxCount, maxDepth, normalize);
            return result;
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Takes a list of input symbols and returns a list of nodes with 
        /// mark value set.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> GetCircleRandomWalkAlpha(List<string> symbols, int maxCount, int maxDepth, bool normalize)
        {
            List<Node> nodes = StringsToNodes(symbols);
            return RW_AllAllowed(nodes, maxCount, maxDepth, normalize);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Takes a list of input symbols and returns a list of nodes with 
        /// mark value set.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> GetCircleRandomWalkBeta(string symbol, int maxCount, int maxDepth, bool normalize)
        {
            List<string> symbols = new List<string>();
            symbols.Add(symbol);
            return GetCircleRandomWalk(symbols, maxCount, maxDepth, normalize);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Takes a list of input symbols and returns a list of nodes with 
        /// mark value set.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> GetCircleRandomWalkBeta(List<string> symbols, int maxCount, int maxDepth, bool normalize)
        {
            return GetCircleRandomWalk(symbols, maxCount, maxDepth, normalize);
        }




        ///---------------------------------------------------------------------
        /// <summary>
        /// Wrapper function.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> GetCircle_RW_SelfAvoiding(Node n, int maxCount, int maxDepth, bool normalize)
        {
            List<Node> nodes = new List<Node>();

            if (n == null)
                n = m_SemanticNetwork.AddSymbol("???");

            nodes.Add(n);

            List<Node> result = RW_SelfAvoiding(nodes, maxCount, maxDepth, normalize);
//            List<Node> result = GetCircleRandomWalkAlpha(nodes, maxCount, maxDepth);
            return result;
        }


        
        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculates an associative ring around one symbol. As a result a list
        /// of nodes is returned with marks set. 
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> RW_AllAllowed(List<Node> nodesInput, int maxCount, int maxDepth, bool normalize)
        {
            List<Node> result = new List<Node>(m_SemanticNetwork.NodeList.Count);

            m_SemanticNetwork.MarkAllNodes(0);


            foreach (Node startNode in nodesInput)
            {
                int count = 0;
                while (count <= maxCount)
                {
                    count++;

                    Node n = startNode;
                    int depth = 0;

                    while (depth <= maxDepth)
                    {
                        if (n.Mark == 0)
                            result.Add(n);

                        n.Mark++;
                        depth++;

                        // Choose a random neighbour
                        int neighborcount = n.Neighbours.Count;
                        if (neighborcount < 1)
                            continue;

                        int idx = rndObj.Next(neighborcount);
                        n = n.Neighbours[idx];
                    }
                }

            }


            result.Sort(SemanticNetwork.cmpMarksDescending);

            if (normalize)
               MarkToRankDescending(result, maxCount);

            return result;
        }
        
        

        

        ///---------------------------------------------------------------------
        /// <summary>
        /// Self avoiding random walk with stop.
        /// The process is done for a couple of input node (start nodes).
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> RW_SelfAvoiding(List<Node> nodesInput, int maxCount, int maxDepth, bool normalize)
        {
            int[] array = new int[m_SemanticNetwork.NodeList.Count];

            m_SemanticNetwork.MarkAllNodes(0);
            
            //
            // currentRunNumber serves to keep the runs apart.
            // Otherwise a problem would occur that the Mark setting
            // from different runs would disturb and block a run.
            //
            int currentRunNumber = 1; 

            List<Node> buffer = new List<Node>();

            foreach (Node startNode in nodesInput)
            {
                int count = 0;
                while (count < maxCount)
                {
                    count++;
                    int depth = 0;

                    Node n = startNode;

                    while (depth <= maxDepth)
                    {
                        depth++;

                        if (n.Mark == currentRunNumber)
                            break;

                        n.Mark = currentRunNumber;

                        int value = array[n.Index]++;

                        // Choose a random neighbour
                        int neighborcount = n.Neighbours.Count;
                        if (neighborcount < 1)
                            continue;

                        int idx = rndObj.Next(neighborcount);
                        n = n.Neighbours[idx];
                    }

                    currentRunNumber++;
                }
            }




            List<Node> result = new List<Node>(m_SemanticNetwork.NodeList.Count);
            foreach (Node n in m_SemanticNetwork.NodeList)
            {
                int value = array[n.Index];
                if (value > 0)
                {
                    result.Add(n);
                    n.Mark = value;
                }
            }

            result.Sort(SemanticNetwork.cmpMarksDescending);

            if (normalize)
                MarkToRankDescending(result, maxCount);
            else
            {
                int maxLog = (int)Math.Round(Math.Log(maxCount));
                foreach (Node n in result)
                    n.Mark = maxLog - (int)Math.Round(Math.Log(n.Mark));
            }

            return result;
        }









        public void MarkToRankAscending(List<Node> nodes, int max)
        {
            int count = nodes.Count;

            if (count == 0)
                return;

            double inverse = 1 / (double)count;

            for (int i = 0; i < count; i++)
                nodes[i].Mark = (int)Math.Round(max * i * inverse);
        }



        public void MarkToRankDescending(List<Node> nodes, int max)
        {
            int count = nodes.Count;

            if (count == 0)
                return;

            double inverse = 1 / (double)count;

            for (int i = 0; i < count; i++)
                nodes[i].Mark = (int)Math.Round(max * (1 - i * inverse));
        }


















        ///---------------------------------------------------------------------
        /// <summary>
        /// Divide a list of symbols into clusters.
        /// For each cluster, a list of nodes is returned.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<List<Node>> Clustering(List<string> symbols, int maxCount, int maxDepth, int countOfClusters)
        {
            List<Node> nodesInput = StringsToNodes(symbols);

            List<List<Node>> result = new List<List<Node>>();
            if (countOfClusters > nodesInput.Count)
                return result; // Empty result.


            List<Node> representants = ChooseRepresentants(nodesInput, maxCount, maxDepth, countOfClusters);

            bool endCondition = false;
            int cc = 0;
            while (!endCondition)
            {
                cc++;
                if (cc > 10) // A reasonable condition for avoiding endless loops.
                    endCondition = true;

                List<Dictionary<Node, int>> dictList = new List<Dictionary<Node, int>>();

                for (int i = 0; i < countOfClusters; i++)
                {
                    List<Node> list = GetCircle_RW_SelfAvoiding(representants[i], maxCount, maxDepth, true); // Circle around current representant
                    
                    Dictionary<Node, int> dict = new Dictionary<Node, int>(); // Map each node to its random walk value.
                    for (int j = 0; j < list.Count; j++)
                        dict[list[j]] = list[j].Mark;
                    dictList.Add(dict);
                }

                // Distribute the  input symbols among the clusters.
                result.Clear();
                for (int i = 0; i < countOfClusters; i++)
                    result.Add(new List<Node>());

                // Loop through the input symbols
                for (int i = 0; i < nodesInput.Count; i++)
                {
                    int maxRandomWalkValue = int.MinValue;
                    int maxIndex = 0; // Will become the index of nearest represant. The nearer, the higher the value.

                    Node currentNode = nodesInput[i];
                    for (int j = 0; j < countOfClusters; j++)
                    {
                        Dictionary<Node, int> dictionaryOfRepresentant = dictList[j];
                        int dist;

                        if (dictionaryOfRepresentant.TryGetValue(currentNode, out dist) == false)
                            dist = int.MinValue;

                        if (dist > maxRandomWalkValue)
                        {
                            maxRandomWalkValue = dist;
                            maxIndex = j;
                        }
                    }

                    // Add input symbol to the representant's list.
                    result[maxIndex].Add(currentNode);
                }





                // Calculate new representants
                for (int i = 0; i < countOfClusters; i++)
                {
                    List<Node> nodes = result[i];
                    nodes = RW_SelfAvoiding(nodes, maxCount, maxDepth, true);
                    if (nodes.Count < 1)
                    {
                        endCondition = true;
                        break;
                    }

                    representants[i] = nodes[0];
                }



            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Choose a representant for each cluster.
        /// </summary>
        ///---------------------------------------------------------------------
        private List<Node> ChooseRepresentants(List<Node> nodesInput, int maxCount, int maxDepth, int countOfClusters)
        {
            List<Node> result = new List<Node>();
            List<Node> available = new List<Node>();

            foreach (Node n in nodesInput)
                available.Add(n);

            Node m = available[0];
            result.Add(m);
            available.Remove(m);

            while (result.Count < countOfClusters)
            {
                List<Node> nodes = this.RW_SelfAvoiding(result, maxCount, maxDepth, true);
                Dictionary<Node, int> dict = new Dictionary<Node, int>();
                foreach (Node n1 in nodes)
                    dict.Add(n1, n1.Mark);

                // Search for the available node with biggest distance to the result set.

                int maxValue = int.MaxValue;
                Node maxDistantNode = null;
                foreach (Node a in available)
                {
                    int d;
                    if (!dict.TryGetValue(a, out d))
                        d = int.MinValue;
                    if (d < maxValue)
                    {
                        maxValue = d;
                        maxDistantNode = a;
                    }
                }

                if (maxDistantNode != null)
                {
                    result.Add(maxDistantNode);
                    available.Remove(maxDistantNode);
                }
                else
                    maxDistantNode = maxDistantNode;
            }

            return result;
        }







        public void Prepare(List<string> symbols, int maxCount, int maxDepth)
        {
            m_CountOfInputSymbols = symbols.Count;

            inputSymbols = StringsToNodes(symbols);
            InitMappingSymbolToValue(inputSymbols, maxCount, maxDepth);
            InitDistanceMatrix(inputSymbols);
        }








        public List<Node> SearchSymbol(List<string> symbols, int maxCount, int maxDepth)
        {
            Prepare(symbols, maxCount, maxDepth);

            List<Node> allNodes = m_SemanticNetwork.NodeList;
            int countOfNodes = allNodes.Count;


            // Loop through all nodes. Evaluate each node
            // as a document. Store the value into .Mark.
            for (int i = 0; i < countOfNodes; i++)
            {
                Node   n = allNodes[i];
                double v = EvaluateNode(n);

                n.Mark = (int)Math.Round(v);
            }

            allNodes.Sort(SemanticNetwork.cmpMarksDescending);
            return allNodes;
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Creates a mapping Node -> Node.Mark.
        /// Necessary since Mark property will be overwritten
        /// by diverse other algorithm steps.
        /// </summary>
        ///---------------------------------------------------------------------
        private Dictionary<Node, int> MarksOfNodes(List<Node> nodes)
        {
            Dictionary<Node, int> result = new Dictionary<Node, int>();

            foreach (Node n in nodes)
                result[n] = n.Mark;

            return result;
        }











        private void InitMappingSymbolToValue(List<Node> inputSymbols, int maxCount, int maxDepth)
        {
            // dictionaries can map each symbol to a distance to one of the
            // input symbols, represented by the index.

            m_DictionariesForRanks = new List<Dictionary<Node, double>>();
            m_DictionariesForRanks2 = new List<Dictionary<Node, int>>();
            m_Alpha = new List<List<NodeAndMark>>();
            m_DictWordToEvaluationVector = new Dictionary<string, EvaluationVector>();

            int countOfInputSymbols = inputSymbols.Count;

            for (int i = 0; i < inputSymbols.Count; i++)
            {
                Node n = inputSymbols[i];

                //--------------------------------------------------------------
                // Random walk circle around current input node.
                //--------------------------------------------------------------

                // Circle of nodes around an input symbol.
                List<Node> list = GetCircle_RW_SelfAvoiding(n, maxCount, maxDepth, false);

                // Create a dictionary Node-to-Mark mapping for the current input symbol.
                Dictionary<Node, int> dict = MarksOfNodes(list);



                int maxLog = (int)Math.Round(Math.Log(maxCount));


                List<NodeAndMark> nodesAndMarks = new List<NodeAndMark>();
                foreach (Node m in list)
                {
                    NodeAndMark p = new NodeAndMark();
                    p.Node = m;
                    p.Mark = m.Mark;
//                    p.Mark = (int)Math.Round(Math.Log(m.Mark));
                    nodesAndMarks.Add(p);
                }

                nodesAndMarks.Sort(NodeAndMark.CmpNodeAndMarkDesc);
//                nodesAndMarks.Sort(NodeAndMark.CmpNodeAndMarkAsc);

                // In nodesAndMarks are now the Nodes that are related to
                // the current input symbol, along with the Mark values.
                // The elements are sorted by Mark descending.
                











                Dictionary<Node, double> dictForRank = new Dictionary<Node, double>();
                int countOfPairs = nodesAndMarks.Count;
                for (int j = 0; j < countOfPairs; j++) // j=i?
                {
                    NodeAndMark p = nodesAndMarks[j];
                    Node m = p.Node;
                    double fraction = j / (double)countOfPairs; // Map to interval [0,1]
                    dictForRank.Add(m, fraction);

                    string s = m.ToString();
                    EvaluationVector vec;
                    if (!m_DictWordToEvaluationVector.TryGetValue(s, out vec))
                    {
                        vec = new EvaluationVector(countOfInputSymbols, maxDepth);
                        m_DictWordToEvaluationVector[s] = vec;
                    }

                    vec[i]         = p.Mark;
                    vec.Symbols[i] = p.ToString();
                }

                m_DictionariesForRanks.Add(dictForRank);




                m_DictionariesForRanks2.Add(dict);
                m_Alpha.Add(nodesAndMarks);
            }
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// 
        /// Initialize a matrix for the pairs of input symbols,
        /// normed to 0..1.
        /// 
        /// The matrix is stored in m_PairDistances.
        /// 
        /// </summary>
        ///---------------------------------------------------------------------
        private void InitDistanceMatrix(List<Node> inputSymbols)
        {



            // Determine count of pairs
            int countOfPairs = 0;
            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = i + 1; j < m_CountOfInputSymbols;  j++)
                    countOfPairs++;





            // Create matrix space
            m_PairDistances_0_1 = new double[countOfPairs];


            // All combinations of input symbols.
            int idx = 0;
            double max = double.MinValue;
            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = i + 1; j < m_CountOfInputSymbols; j++)
                {
                    double dist = DetermineDistance(inputSymbols, i, j);
                    m_PairDistances_0_1[idx++] = dist;

                    if (dist > max)
                        max = dist;
                }




            // Norming so that max dist == 1
            if (max > 0)
               for (int i = 0; i < countOfPairs; i++)
                    m_PairDistances_0_1[i] /= max;
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Determine the semantic distance between two input symbols
        /// normed to 0..1. This is done by taking the random walk map
        /// of the Node i and retrieving the entry for node j.
        /// </summary>
        ///---------------------------------------------------------------------
        private double DetermineDistance(List<Node> inputSymbols, int i, int j)
        {
            // The dictionary maps many Nodes to ranks, where rank is
            // meant here as a distance, normed to 0..1. That means, 1
            // is farthest arrays.
            Dictionary<Node, double> dict = m_DictionariesForRanks[i];
            Node n_i = inputSymbols[i];
            Node n_j = inputSymbols[j];

            double value_i; // Not needed. Only for debugging reasons. Should always be 0.0
            double value_j;

            if (dict.TryGetValue(n_i, out value_i) == false)
                value_i = 1;

            if (dict.TryGetValue(n_j, out value_j) == false)
                value_j = 1;

            double result = value_j;
            return result;
        }
        




        /*
        private void InitDistanceMatrix(List<Node> inputSymbols)
        {
            // Determine count of pairs
            int countOfPairs = 0;
            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = i + 1; j < m_CountOfInputSymbols; j++)
                    countOfPairs++;

            // Create matrix space
            m_PairDistances = new double[countOfPairs];

            int idx = 0;
            for (int i = 0; i < m_CountOfInputSymbols; i++)
                for (int j = i + 1; j < m_CountOfInputSymbols; j++)
                {
                    double dist = DetermineDistance(inputSymbols, i, j);
                    m_PairDistances[idx++] = dist;
                }
        }

        private double DetermineDistance(List<Node> inputSymbols, int i, int j)
        {
            List<PairOfNodeAndMark> list1 = m_Alpha[i];
            List<PairOfNodeAndMark> list2 = m_Alpha[j];

            int countOfAllRelevantNodes = 0;
            Dictionary<Node, int> dict = new Dictionary<Node, int>();

            foreach (PairOfNodeAndMark p in list1)
                if (!dict.ContainsKey(p.Node))
                {
                    dict.Add(p.Node, 1);
                    countOfAllRelevantNodes++;
                }

            foreach (PairOfNodeAndMark p in list2)
                if (!dict.ContainsKey(p.Node))
                {
                    dict.Add(p.Node, 1);
                    countOfAllRelevantNodes++;
                }

            double result;
            if (countOfAllRelevantNodes > 0)
                result = list2.Count / (double)countOfAllRelevantNodes;
            else
                result = 0.0;

            return result;
        }

*/










        ///---------------------------------------------------------------------
        /// <summary>
        /// For one node, sum up values for all pairs of 
        /// input symbols of the query.
        /// </summary>
        ///---------------------------------------------------------------------
        public double EvaluateNode(Node n)
        {
            double sum = 0.0;
//            int sum = 0;

            int currentPair = 0;
            for (int i = 0; i < m_CountOfInputSymbols; i++)
            {
                int dist_i;

                if (m_DictionariesForRanks2[i].TryGetValue(n, out dist_i) == false)
                    dist_i = 0; // Assume maximal distance

                // Special case: only one input symbol.
                // The next loop is then not being run.
                if (m_CountOfInputSymbols == 1)
                    sum += dist_i;



                for (int j = i + 1; j < m_CountOfInputSymbols; j++)
                {
                    int dist_j;

                    if (m_DictionariesForRanks2[j].TryGetValue(n, out dist_j) == false)
                        dist_j = 0; // Assume maximal distance

                    int maxDist = Math.Max(dist_i, dist_j);
                    int minDist = Math.Min(dist_i, dist_j);

                    // Distance between Node i and Node j
                    double ratio = m_PairDistances_0_1[currentPair];
                   
                    sum += maxDist * ratio + minDist * (1 - ratio);
                    currentPair++;
                }
            }

            return sum;
        }
    }
}
