//------------------------------------------------------------------------------
// Purpose: A concept of semantic distance by the number of the edges that
//          occur on the path between two nodes.
// Date   : 13.Sep.2008
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace Fusym
{
    public class Propagation
    {
        public Node FromNode;
        public int Index;
        public int Value;
        public int Level;
    }


    public class SemanticDistanceByEdgeNumber
    {
        private Node target;
        private int m_MaxCircleValue;
        private Queue<Propagation> m_Propagations = new Queue<Propagation>();
        private SemanticNetwork m_SemanticNetwork;
        private Dictionary<Node,int> m_AlreadyVisited = new Dictionary<Node,int>();

        ///---------------------------------------------------------------------
        /// <summary>
        /// Constructor.
        /// </summary>
        ///---------------------------------------------------------------------
        public SemanticDistanceByEdgeNumber(SemanticNetwork cm)
        {
            m_SemanticNetwork = cm;
        }
        
        
        public void CreateAscendingValueAssignment(Node start, Node dest)
        {
            m_Propagations.Clear();
        
            Node d = start;
            target = dest;
            d.Mark = 0;
            target.Mark = int.MaxValue;
            
            m_AlreadyVisited.Clear();
            m_AlreadyVisited.Add(start,0);
            d.InitEdgeValues(int.MaxValue);
            
            // Initialize the queue with all edges that go out of the start node
            for (int i = 0; i < d.Neighbours.Count; i++)
            {
                Propagation p = new Propagation(); // Eigene Verwaltung implementieren! Nachtrag: wohl eigene Liste, damit nicht New aufgerufen werden mu.
                p.FromNode = d;
                p.Index    = i;
                p.Value    = d.Neighbours.Count;
                p.Level    = 0;
                
                m_Propagations.Enqueue(p);
            }
            
            
                
            while (m_Propagations.Count != 0)
            {
                Propagation p = m_Propagations.Dequeue();
                Propagate(p);
            }
        }
        
        

        
        
        ///---------------------------------------------------------------------
        /// <summary>
        /// Find the path from one node to another.
        /// </summary>
        ///---------------------------------------------------------------------
        public bool FindPath(string start, string end, List<Node> lis)
        {
            Node startNode = m_SemanticNetwork.GetNode(start);
            Node destNode  = m_SemanticNetwork.GetNode(end);
            
            if ((startNode == null) || (destNode == null))
               return false;
               
               
            // For the sake of security, check if there's _any_ path.
            if (m_SemanticNetwork.FindPath(start,end,lis) == false)
               return false;   
               
            m_SemanticNetwork.MarkAllNodes(int.MaxValue);  
            
            CreateAscendingValueAssignment(startNode, destNode);

            
            Node n1 = startNode;
            Node n2 = destNode;
            
            lis.Clear();
            lis.Add(n2);
            
            Node current = n2;            
            while (current != n1)
            {
                Node nextNode = null;              
            
                // Search for a neighbour node with a lower sum (Mark + Neighbourcount)
                foreach (Node n in current.Neighbours)
                {
                    if (
                        (n.Mark < int.MaxValue) && 
                        (n.Mark + n.Neighbours.Count <= current.Mark)
                       )
                    {
                        if (
                            (nextNode == null) || 
                            (n.Mark + n.Neighbours.Count < nextNode.Mark) // Event better
                           )
                           nextNode = n;
                    }
                }
                
                if (nextNode != null)
                {
                    current = nextNode;
                    lis.Add(current);
                }
            }
            
            lis.Reverse();
            
            return true;
        }
        
        
        
        
        
        
        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculate the semantic distance between two symbols.
        /// 
        /// </summary>
        /// <param name="symbol1">One symbol</param>
        /// <param name="symbol2">Another symbol</param>
        /// <returns>The length of the path in the graph between the symbols, -1 if there is no path</returns>
        ///---------------------------------------------------------------------
        public int SemanticDistance(string symbol1, string symbol2)
        {
            List<Node> lis = new List<Node>();

            if (FindPath(symbol1, symbol2, lis))
               {
                  int result = 0;
                  
                  for (int i = 1; i < lis.Count - 1; i++)
                      result += lis[i].Neighbours.Count;
                      
                  return result;
               }
            else
               return -1;
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Retrieve the environment of a node, all nodes that
        /// are within the circle of a certain radius.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> Circle(string symbol, int maxCircleValue)
        {
            List<Node> list = new List<Node>();

            Node start = m_SemanticNetwork.GetNode(symbol);

            if (start == null)
                return list;
            
            m_MaxCircleValue = maxCircleValue;
            m_SemanticNetwork.MarkAllNodes(int.MaxValue);
        
            m_Propagations.Clear();
        
            Node d = start;
            d.Mark = 0;
            d.InitEdgeValues(int.MaxValue); // Test, 25.2.2009
            
            // Initialize the queue with all edges that go out of the start node
            for (int i = 0; i < d.Neighbours.Count; i++)
            {
                Node e = d.Neighbours[i];
            
                Propagation p = new Propagation(); // Eigene Verwaltung zu-tun
                p.FromNode = d;
                p.Index    = i;
                p.Value    = d.Neighbours.Count;
                p.Level    = 0;
                
                m_Propagations.Enqueue(p);
            }
            
            
                
            while (m_Propagations.Count != 0)
            {
                Propagation p = m_Propagations.Dequeue();
                Propagate2(p, list);
            }
            
            foreach (Node n in m_SemanticNetwork.NodeList)
            {
                Node d2 = n;
                if (d2.Mark < int.MaxValue)
                {
                     list.Add(d2);
//Console.WriteLine(String.Format("{0:D5}",d2.TheNode.Mark) +' ' + d2.TheNode.ToString() + ' ' + d2.TheNode.Mark.ToString());
                }
            }

            return list;
        }  
        
        
        ///---------------------------------------------------------------------
        /// <summary>
        /// Suggest a new value to an edge.
        /// </summary>
        ///---------------------------------------------------------------------
        private void Propagate(Propagation p)
        {
            Node e    = p.FromNode;
            int value = p.Value;
            int level = p.Level;
        
            if (value >= e.EdgeValue[p.Index]) // No improvement for the edge
               return; 
               
            if (value >= target.Mark)
               return; // Mu ich nochmal drber nachdenken zu-tun

           
            e.EdgeValue[p.Index] = value; // Decrease the value of the edge  

            Node toNode = e.Neighbours[p.Index];
            
            if (! m_AlreadyVisited.ContainsKey(toNode))
            {
               toNode.InitEdgeValues(int.MaxValue); // Fraglich
               m_AlreadyVisited.Add(toNode,0);
            }
            
            if (value < toNode.Mark)
               {
                    toNode.Mark = value; // Decrease the value of the target node  
                    
                    //foreach (DistEdge e2 in e.ToNode.Neighbours)
                    for (int i = 0; i < toNode.Neighbours.Count; i++)
                    {
                        Propagation p2 = new Propagation(); // Eigene Verwaltung zu-tun
                        p2.FromNode = toNode;
                        p2.Index    = i;
                        p2.Value    = toNode.Mark + toNode.Neighbours.Count;
                        p2.Level    = level + 1;
                        
                        m_Propagations.Enqueue(p2);                        
                    }
                        
               }
        }
        
        
        
        private void Propagate2(Propagation p, List<Node> list)
        {
            Node e    = p.FromNode;
            int value = p.Value;
            int level = p.Level;
        
            if (value >= e.EdgeValue[p.Index]) // No improvement for the edge
               return; 
               
            if (level > m_MaxCircleValue)
               return; // Fraglich, ob das die beste Methode ist

            e.EdgeValue[p.Index] = value;            

            Node toNode = e.Neighbours[p.Index];

            if (! m_AlreadyVisited.ContainsKey(toNode))
            {
               toNode.InitEdgeValues(int.MaxValue); // Fraglich
               m_AlreadyVisited.Add(toNode,0);
            }
            
            if (value < toNode.Mark)
               {
                    toNode.Mark = value; // Decrease the value of the target node  
                    
                    for (int i = 0; i < toNode.Neighbours.Count; i++)
                    {
                        Propagation p2 = new Propagation(); // zu-tun
                        p2.FromNode = toNode;
                        p2.Index    = i;
                        p2.Value    = toNode.Mark + toNode.Neighbours.Count;
                        p2.Level    = level + 1;
                        
                        m_Propagations.Enqueue(p2);                        
                    }
                        
               }
        }
              

    }
    
    
}
