//------------------------------------------------------------------------------
// Purpose: Managing associations in Fusym technology.
// Date   : 26.Apr.2006
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Collections;

namespace Fusym
{
    ///-------------------------------------------------------------------------
    /// <summary>
    /// Represents the associative network
    /// </summary>
    ///-------------------------------------------------------------------------
    public class SemanticNetwork : IDisposable
    {
        Dictionary<string,Node> symbolHash  = new Dictionary<string,Node>();
        
        List<Node> m_NodeList = new List<Node>(200000);
        private Encoding m_Encoding;

        public Encoding TheEncoding
        {
            get { return m_Encoding; }
            set { m_Encoding = value; }
        }

        public List<Node> NodeList
        {
            get { 
                 return m_NodeList; 
                }
            set { m_NodeList = value; }
        }

        


        public void Dispose()
        {
            return;
        }


        public SemanticNetwork(Encoding encoding)
        {
            m_Encoding = encoding;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Add an edge between two Nodes in both directions.
        /// </summary>
        ///---------------------------------------------------------------------
        public void AddPair(Node sym1, Node sym2)
        {

            if (!sym1.Neighbours.Contains(sym2))
                sym1.Neighbours.Add(sym2);


            if (!sym2.Neighbours.Contains(sym1))
                sym2.Neighbours.Add(sym1);

            /*
            sym1.AddNeighbour(sym2);
            sym2.AddNeighbour(sym1);
             */
        }

        public void RemovePair(string symbol1, string symbol2)
        {
            Node sym1 = GetNode(symbol1);
            Node sym2 = GetNode(symbol2);

            while (sym1.Neighbours.Contains(sym2))
                sym1.Neighbours.Remove(sym2);

            while (sym2.Neighbours.Contains(sym1))
                sym2.Neighbours.Remove(sym1);
        }

        public void RemovePair(Node sym1, Node sym2)
        {
            while (sym1.Neighbours.Contains(sym2))
                sym1.Neighbours.Remove(sym2);

            while (sym2.Neighbours.Contains(sym1))
                sym2.Neighbours.Remove(sym1);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Isolate a node from its neighbors. It's virtually
        /// like deleting the node.
        /// </summary>
        ///---------------------------------------------------------------------
        public void Isolate(Node n)
        {
            IList<Node> neighbors = n.Neighbours;
            int countM1 = neighbors.Count - 1;

            for (int i = countM1; i >= 0; i--)
            {
                Node other = neighbors[i];
                RemovePair(n, other);
            }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Add a list of nodes to the graph.
        /// Each pair (order is irrelevant) builds up an edge.
        /// </summary>
        /// <param name="cluster">A list of symbols (strings).</param>
        ///---------------------------------------------------------------------
        public void AddCluster(List<Node> cluster)
        {
            if (cluster.Count <= 0)
               return;

           for (int i = 0; i < cluster.Count - 1; i++)
               for (int j = i + 1; j < cluster.Count; j++)
                       AddPair(cluster[i], cluster[j]);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Makes a symbol known.
        /// </summary>
        /// <param name="symbol">An uppercase string without spaces, a symbol</param>
        /// <returns>The Node object of the symbol.</returns>
        ///---------------------------------------------------------------------
        public Node AddSymbol(string symbol)
        {
            Node result;

            // Already stored? 
            if (symbolHash.TryGetValue(symbol, out result))
                return result;


            result = new Node(symbol);
            symbolHash[symbol] = result;

            result.Index = m_NodeList.Count;
            m_NodeList.Add(result);

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Gives a Node instance for a symbol string.
        /// </summary>
        /// <param name="symbol">String representation of the symbol</param>
        /// <returns>A Node instance, if there is one, null otherwise</returns>
        ///---------------------------------------------------------------------
        public Node GetNode(string symbol)
        {
            Node result;

            symbol = symbol.ToUpper();

            if (symbolHash.TryGetValue(symbol, out result))
                return result;
            else
                return null;
        }


        public Node GetNodeTolerant(string symbol)
        {
            Node result;

            symbol = symbol.ToUpper();

            if (symbolHash.TryGetValue(symbol, out result))
                return result;
            else
            {
                string s = symbol.Replace("SS", "");
                symbolHash.TryGetValue(s, out result);
                return result;
            }
        }






        ///---------------------------------------------------------------------
        /// <summary>
        /// Check if a symbol is known and has at least one neighbour
        /// </summary>
        /// <param name="symbol">String representation of the symbol</param>
        /// <returns></returns>
        ///---------------------------------------------------------------------
        public bool HasAssociations(string symbol)
        {
            Node n;

            symbol = symbol.ToUpper();
           
            if (symbolHash.TryGetValue(symbol, out n) == false)
                n = null;

            if (n == null)
               return false;
            else
               return n.GetFirstNeighbour() != null;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Loads a cluster file and integrates its associations into the graph
        /// </summary>
        /// <param name="filename">A filename</param>
        ///---------------------------------------------------------------------
        public void LoadFromFile(string filename)
        {
            List<Node> cluster = new List<Node>();
            Regex regexSpace = new Regex(@"^\s*$", RegexOptions.Compiled);
            Regex regexNoSpace = new Regex(@"^\s*\S+\s*$", RegexOptions.Compiled);


            StreamReader sr = new StreamReader(new BufferedStream(new FileStream(filename, FileMode.Open), 10000000), m_Encoding);

            string line = "";
            while ((line = sr.ReadLine()) != null)
            {
                if (regexNoSpace.IsMatch(line))
                {
                    Node sym = AddSymbol(line.ToUpper().Trim());
                    cluster.Add(sym);
                }
                else
                    if (regexSpace.IsMatch(line))
                    {
                        AddCluster(cluster);
                        cluster.Clear();
                    }
                /* 
                else
                  unknown line type.
                */
            }

            AddCluster(cluster);
            cluster.Clear();

            sr.Close();
        }


        ///----------------------------------------------------------------------
        /// <summary>
        /// Load clusters and create associations of symbols.
        /// The source of clusters implements the IClusterProvider
        /// interface.
        /// </summary>
        ///----------------------------------------------------------------------
        public void LoadFromClusterProvider(IClusterProvider clusterProvider)
        {
            if (clusterProvider == null)
               return;

            List<string> stringList = clusterProvider.GetFirstCluster();
            List<Node> cluster      = new List<Node>();
            while (stringList != null)
            {
                cluster.Clear();
                for (int i = 0; i < stringList.Count; i++)
                {
                    Node sym = AddSymbol(stringList[i]);
                    cluster.Add(sym);
                }


                AddCluster(cluster);
                stringList = clusterProvider.GetNextCluster();
            }
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Set the Mark property of all Node objects to value.
        /// </summary>
        /// <param name="value">The int value</param>
        ///---------------------------------------------------------------------
        public void MarkAllNodes(int value)
        {
            for (int i = 0; i < m_NodeList.Count; i++)
            {
                Node n = m_NodeList[i];
                n.Mark = value;
            }
        }




        ///---------------------------------------------------------------------
        /// <summary>
        /// Find a path from one symbol to another in the graph.
        /// </summary>
        /// <param name="start">A symbol string</param>
        /// <param name="end">Another symbol string</param>
        /// <param name="lis">List that will be filled with a list of Node objects on the path</param>
        /// <returns>true if a path could be found, false otherwise</returns>
        ///---------------------------------------------------------------------
        public bool FindPath(string start, string end, List<Node> list)
        {
            Node startNode = GetNode(start);
            Node endNode   = GetNode(end);
            return FindPath(startNode, endNode, list);
        }

        public bool FindPath(Node startNode, Node endNode, List<Node> list)
        {
            Node current;
            bool found = false;
            int nextMark;
            list.Clear();

            if ((startNode == null) || (endNode == null))
                return false;

            Queue q = new Queue();
            found = false;
            MarkAllNodes(-1);
            startNode.Mark = 0;
            q.Enqueue(startNode);
            while (q.Count != 0)
            {
                current = (Node)q.Dequeue();
                if (current == endNode)
                {
                    found = true;
                    break;
                }

                // Mark all m_Neighbours of current
                nextMark = current.Mark + 1;
                Node m = current.GetFirstNeighbour();
                while (m != null)
                {
                    if (m.Mark < 0) // Not already visited
                       {
                        m.Mark = nextMark;
                        q.Enqueue(m);
                       }
                    m = current.GetNextNeighbour();
                }
            }



            if (!found)
               return false;



            // Now step back from endNode to startNode
            current = endNode; // To make the compiler happy
            list.Add(endNode);
            nextMark = endNode.Mark - 1;
            while (current != startNode)
            {
                // Search for a neighbour of currentnode with Mark == nextMark
                Node m = current.GetFirstNeighbour();
                while (m != null)
                {
                    if (m.Mark == nextMark) // Not already visited
                       {
                        current = m;
                        list.Add(current);
                        nextMark--;
                        break;
                       }
                    m = current.GetNextNeighbour();
                }
            }

            list.Reverse();
            return true;
        }





        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculates an associative ring around one symbol.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<Node> Circle(string symbol, int maxCount, int maxDepth)
        {
            List<Node> list = new List<Node>();

            int count = 0;
            int depth = 0;


            // If the symbol is not known, return
            // a list with one node as the circle.
            Node n = GetNode(symbol);
            if (n == null)
            {
                Node newNode = AddSymbol(symbol.ToUpper());
                newNode.Mark = 0;
                list.Add(newNode);
                return list;
            }




            Queue<Node> queue = new Queue<Node>();

            MarkAllNodes(-1);
            n.Mark = 0;
            queue.Enqueue(n);

            while (queue.Count != 0)
            {
                n = queue.Dequeue();

                depth = n.Mark;
                count++;

                if (
                    (count <= maxCount) && 
                    (depth++ <= maxDepth)
                   )
                   list.Add(n);
                else
                   return list;

                // Iteration through all m_Neighbours of the current node n

                foreach (Node m in n.Neighbours)
                {
                    if (m.Mark < 0) // Not already marked?
                    {
                        m.Mark = depth;
                        queue.Enqueue(m);
                    }
                }
            }

            return list;
        }





        ///---------------------------------------------------------------------
        /// <summary>
        /// Calculate the semantic distance between two symbols.
        /// 
        /// Man kann hier optimieren; der Rueckweg ist uninteressant.
        /// </summary>
        ///---------------------------------------------------------------------
        public int SemanticDistance(string symbol1, string symbol2)
        {
            List<Node> lis = new List<Node>();

            if (FindPath(symbol1, symbol2, lis))
               return lis.Count - 1;
            else
               return -1;
        }




        public int SemanticDistanceByCloudMatching(string symbol1, string symbol2, int maxCount, int maxDepth)
        {
            CloudMatching cm = new CloudMatching(this);
            int result = cm.SemanticDistance(symbol1, symbol2, maxCount, maxDepth);
            return result;
        }


        // Gehrt nicht hier hinein zu-tun
        public IDocumentEvaluator NewDocumentEvaluator(
                                                      List<string> inputSymbols, 
                                                      int depth, 
                                                      int maxCount, 
                                                      EvaluationMethod evaluationMethod,
                                                      bool fullTextSearch
                                                     )
        {
            if (evaluationMethod == EvaluationMethod.RandomWalk)
                return new DocumentEvaluatorRandomWalk(this, inputSymbols, depth, maxCount);
            else
                return new DocumentEvaluator(
                                         this, 
                                         inputSymbols, 
                                         depth, 
                                         maxCount, 
                                         evaluationMethod,
                                         fullTextSearch
                                        );
        }


        public static int cmpMarksAscending(Node a, Node b)
        {
            return a.Mark.CompareTo(b.Mark);
        }

        public static int cmpMarksDescending(Node a, Node b)
        {
            return -a.Mark.CompareTo(b.Mark);
        }

        public void SortByMark()
        {
            NodeList.Sort(cmpMarksAscending);
        }
    }
}
