// Joachim Pimiskern, 26.6.2010
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Text.RegularExpressions;

namespace Fusym
{
    public class SymbolAsBitsManager
    {
        private static string m_RegexDateString = @"^(?<bits>[01]+)\s(?<symbol>[\S]+).*$";
        private static Regex m_RegexSymbolBits = new Regex(m_RegexDateString, RegexOptions.Compiled);  





        private SemanticNetwork m_SemanticNetwork;
        public  int BitsPerSymbol;
        private Dictionary<string, SymbolAsBits> m_Dict = new Dictionary<string, SymbolAsBits>();
        private Random m_Rnd = new Random();
        private List<SymbolAsBits> m_List = new List<SymbolAsBits>();
        private int m_Count;
        private int[] m_Collector;
        private Dictionary<string, SymbolAsBits> m_CollisionAvoidance = new Dictionary<string, SymbolAsBits>();

        public SymbolAsBitsManager(SemanticNetwork sn, int bitsPerSymbol)
        {
            m_SemanticNetwork = sn;
            BitsPerSymbol   = bitsPerSymbol;
            m_Collector = new int[bitsPerSymbol];
            Load();
            Init();
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// For every node in the semantic network:
        /// invent a bit pattern, if it doesn't already exist.
        /// It might have been loaded from a file.
        /// </summary>
        ///---------------------------------------------------------------------
        private void Init()
        {
            List<Node> nodes = m_SemanticNetwork.NodeList;
            for (int i = 0; i < nodes.Count; i++)
            {
                Node n = nodes[i];
                string symbolAsString = n.ToString();

                if (!m_Dict.ContainsKey(symbolAsString))
                {
                    SymbolAsBits sb = new SymbolAsBits(n, BitsPerSymbol);
                    m_Dict.Add(symbolAsString, sb);
                    m_List.Add(sb);
                }
            }

            m_Count = m_List.Count;
        }


        public double ErrorAvg()
        {
            double sum = 0;

            int count = m_List.Count;

            for (int i = 0; i < count; i++)
            {
                SymbolAsBits sb = m_List[i];
                int sum2 = 0;
                int count2 = 0;

                Node n = sb.Node;
                Node m = n.GetFirstNeighbour();
                while (m != null)
                {
                    count2++;
                    string s = m.ToString();
                    SymbolAsBits s2;
                    if (m_Dict.ContainsKey(s))
                    {
                        s2 = m_Dict[s];
                        sum2 += sb.DistanceTo(s2);
                    }


                    m = n.GetNextNeighbour();
                }

                if (count2 > 0)
                    sum += sum2 / (double)count2;
            }

            if (count > 0)
                return sum / (double)count;
            else
                return 0;
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Hamming distance between two symbols.
        /// </summary>
        ///---------------------------------------------------------------------
        public int Distance(string s1, string s2)
        {
            SymbolAsBits sb1, sb2;
            s1 = s1.ToUpper();
            s2 = s2.ToUpper();

            if (m_Dict.ContainsKey(s1))
                sb1 = m_Dict[s1];
            else
                return BitsPerSymbol;

            if (m_Dict.ContainsKey(s2))
                sb2 = m_Dict[s2];
            else
                return BitsPerSymbol;

            int result = sb1.DistanceTo(sb2);
            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// One training step.
        /// </summary>
        ///---------------------------------------------------------------------
        public string TrainingStep()
        {
            Array.Clear(m_Collector, 0, BitsPerSymbol);


            // Choose a random bitvector of a symbol.
            int idx1 = m_Rnd.Next(m_List.Count);
            SymbolAsBits s1 = m_List[idx1];

            //
            // This block adds up all bit vectors of the
            // neighbors of the current symbol.
            //
            int countOfNeighbours = 0; // Neighbors of chosen symbol.
            Node n1 = s1.Node;
            Node n2 = n1.GetFirstNeighbour();
            while (n2 != null) // For all neighbours
            {
                countOfNeighbours++;

                string s = n2.ToString();
                SymbolAsBits s2 = m_Dict[s];

                for (int i = 0; i < BitsPerSymbol; i++)
                    m_Collector[i] += s2.Bits[i];

                n2 = n1.GetNextNeighbour();
            }


            // No neighbor found.
            if (countOfNeighbours == 0)
                return "";

            double inverseNeighborCount = 1.0 / (double)countOfNeighbours;


            /*

            // If there's an odd number of neighbours, add the chosen middle symbol's bits.
            if (countOfNeighbours % 2 != 0)
            {
                countOfNeighbours++; // Now it's even.
                for (int i = 0; i < m_BitsPerSymbol; i++)
                    m_Collector[i] += s1.Bits[i];
            }

            */



            countOfNeighbours++;
            for (int i = 0; i < BitsPerSymbol; i++)
                m_Collector[i] += s1.Bits[i];


            // Half of the number of neighbours.
            int threshold = countOfNeighbours >> 1;

            if (countOfNeighbours % 2 != 0)
            {
                if (m_Rnd.NextDouble() > 0.5) // Introduces entropy.
                    threshold++;
            }

            






            for (int i = 0; i < BitsPerSymbol; i++)
            {
                int bit = m_Collector[i];
                if (bit > threshold) // More neighbors have a 1 at this digit than a 0.
                    s1.Bits[i] = 1;
                else
                if (bit < threshold) // More neighbors have a 0 at this digit than a 1.
                    s1.Bits[i] = 0;  
                else
                    // Entropy. Introduce random bits
                    s1.Bits[i] = (byte)m_Rnd.Next(2); // 0 or 1
            }





            //
            // If a neighbor has the same code, increase the
            // distance to it by flipping a bit.
            //
            n2 = n1.GetFirstNeighbour();
            while (n2 != null) // For all neighbours
            {
                string s = n2.ToString();
                SymbolAsBits s2 = m_Dict[s];

                if (s1.DistanceTo(s2) == 0)
                {
  //                if (m_Rnd.Next(2) == 0) // with 50% probability
                  if (m_Rnd.NextDouble() <= inverseNeighborCount)
                     s1.IncreaseDistanceTo(s2);
                  else
                     s2.IncreaseDistanceTo(s1);
                }

                n2 = n1.GetNextNeighbour();
            }



            string result = s1.ToString();
            return result;
        }

        
        public void AvoidCollisions()
        {
            int[] indexes = new int[m_Count];
            for (int i = 0; i < m_Count; i++)
                indexes[i] = i;

            for (int i = 0; i < m_Count; i++)
            {
                int otherIndex = m_Rnd.Next(0, m_Count);
                int h = indexes[otherIndex];
                indexes[otherIndex] = indexes[i];
                indexes[i] = h;
            }



            m_CollisionAvoidance.Clear();
            for (int i = 0; i < m_Count; i++)
            {
                SymbolAsBits sb = m_List[indexes[i]];
                string key = sb.ToStringOnlyBits();

                while (m_CollisionAvoidance.ContainsKey(key))
                {
                    SymbolAsBits other = m_CollisionAvoidance[key];
                    sb.IncreaseDistanceTo(other);
                    key = sb.ToStringOnlyBits(); // One bit has changed, so determine key again.
                }

                m_CollisionAvoidance.Add(key, sb);
            }
        }
        

        public string Filename()
        {
            string path = Path.GetDirectoryName(Application.ExecutablePath) + "\\database\\";
            string result = path + "symbolbits.txt";
            return result;
        }


        public void Save()
        {
            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < m_List.Count; i++)
            {
                SymbolAsBits s = m_List[i];
                Node n = s.Node;

                sb.AppendLine(s.ToString());
            }



            string fn = Filename();
            SaveToFile(fn, sb.ToString());
        }


        public bool SaveToFile(string filename, string s)
        {
            try
            {
                FileStream fs = new FileStream(filename, FileMode.Create);
                if (!fs.CanWrite)
                {
                    fs.Close();
                    return false;
                }

                StreamWriter writer = new StreamWriter(fs, System.Text.Encoding.GetEncoding(1252));
                writer.WriteLine(s);
                writer.Flush();
                writer.Close();
                fs.Close();
                return true;
            }
            catch
            {
                return false;
            }
        }




        private void Load()
        {
            string fn = Filename();

            if (!File.Exists(fn))
                Save();

            List<string> lines = LoadFromFile(fn);

            for (int i = 0; i < lines.Count; i++)
            {
                string s = lines[i];

                Match m = m_RegexSymbolBits.Match(s);
                if (m.Success)
                {
                    string bits = m.Groups["bits"].Value;
                    string symbol = m.Groups["symbol"].Value.ToUpper();

                    Node n = m_SemanticNetwork.GetNode(symbol);
                    if (n == null)
                        continue;

                    SymbolAsBits sb = new SymbolAsBits(n, BitsPerSymbol);
                    sb.Bits = StringToBits(bits);
                    m_Dict.Add(n.ToString(), sb);
                    m_List.Add(sb);
                }
            }


            m_Count = m_List.Count;
        }

        private byte[] StringToBits(string bits)
        {
            int len = bits.Length;
            byte[] result = new byte[len];
            for (int i = 0; i < len; i++)
            {
                char c = bits[i];
                if (c == '0')
                    result[i] = 0;
                else
                    result[i] = 1;
            }

            return result;
        }





        private List<string> LoadFromFile(string filename)
        {
            List<string> result = new List<string>();

            StreamReader sr = null;

            try
            {
                sr = new StreamReader(
                                      new BufferedStream(new FileStream(filename, FileMode.Open)),
                                      System.Text.Encoding.GetEncoding(1252)
                                     );

                string line = "";

                while ((line = sr.ReadLine()) != null)
                {
                    result.Add(line.Trim());
                }
            }
            finally
            {
                sr.Close();
            }

            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Find bitstring of a symbol. If it doesn't exist, return null.
        /// </summary>
        ///---------------------------------------------------------------------
        public SymbolAsBits Find(string s)
        {
            s = s.Trim().ToUpper();
            if (m_Dict.ContainsKey(s))
                return m_Dict[s];
            else
                return null;
        }


        byte[] Combine(List<SymbolAsBits> list)
        {
            byte[] result = new byte[BitsPerSymbol];
            int count = list.Count;

            for (int j = 0; j < BitsPerSymbol; j++)
            {
                int c0 = 0;
                int c1 = 0;

                for (int i = 0; i < count; i++)
                {
                    byte[] bits = list[i].Bits;
                    if (bits[j] == 0)
                        c0++;
                    else
                        c1++;
                }

                if (c0 > c1)
                    result[j] = 0;
                else
                    result[j] = 1;
            }

            return result;
        }


        public List<string> SearchByString(List<string> querySymbols, bool bGroupChange)
        {
            List<string> result = new List<string>();

            List<SymbolAsBits> list = new List<SymbolAsBits>();


            for (int i = 0; i < querySymbols.Count; i++)
            {
                string s = querySymbols[i];
                SymbolAsBits sb = Find(s);

                if (sb != null)
                    list.Add(sb);
            }


            byte[] data = Combine(list);

            FillStringList(bGroupChange, result, data);
            return result;
        }

        public List<String> SearchByBits(string s, bool bGroupChange)
        {
            List<string> result = new List<string>();

            int min = Math.Min(s.Length, BitsPerSymbol);
            byte[] data = new byte[min];

            for (int i = 0; i < min; i++)
                if (s[i] == '1')
                    data[i] = 1;
                else
                    data[i] = 0;

            FillStringList(bGroupChange, result, data);
            return result;
        }







        private void FillStringList(bool bGroupChange, List<string> result, byte[] data)
        {
            //------------------------------------------------------------------
            // Fill a list of pairs, consisting of
            // SymbolAsBits and the distance to the representation
            // to s.
            //------------------------------------------------------------------
            List<SymbolAsBitsAndDistance> list = new List<SymbolAsBitsAndDistance>();
            for (int i = 0; i < m_List.Count; i++)
            {
                SymbolAsBits sb2 = m_List[i];
                SymbolAsBitsAndDistance sd = new SymbolAsBitsAndDistance();
                sd.SymbolAsBits = sb2;
                sd.Distance = Distance(data, sb2.Bits);
                list.Add(sd);
            }

            list.Sort(cmpSymbolAsBitsAndDistance);


            int oldDistance = 0;
            for (int i = 0; i < list.Count; i++)
            {
                SymbolAsBitsAndDistance sd = list[i];
                int distance = sd.Distance;
                if (bGroupChange && (distance != oldDistance))
                    result.Add("");
                oldDistance = distance;

                string s1 = sd.SymbolAsBits.Node.ToString();
                result.Add(sd.ToString());
            }
        }


        private int cmpSymbolAsBitsAndDistance(SymbolAsBitsAndDistance a, SymbolAsBitsAndDistance b)
        {
            return a.Distance.CompareTo(b.Distance);
        }



        private int Distance(byte[] data, byte[] p)
        {
            int result = 0;
            for (int i = 0; i < data.Length; i++)
            {
                if (data[i] != p[i])
                    result++;
            }


            return result;
        }




        ///---------------------------------------------------------------------
        /// <summary>
        /// Shows the distance distribution with respect to
        /// a given symbol s. Result should by a text respresentation of
        /// a Gaussian curve.
        /// </summary>
        ///---------------------------------------------------------------------
        public List<string> Distribution(string s)
        {
            List<string> result = new List<string>();
            s = s.Trim().ToUpper();

            SymbolAsBits sb;

            if (m_Dict.ContainsKey(s))
                sb = m_Dict[s];
            else
                return result;

            byte[] data = sb.Bits;


            int[] occurrences = new int[BitsPerSymbol];
            for (int i = 0; i < m_Count; i++)
            {
                int diff = m_List[i].DistanceTo(sb);
                occurrences[diff]++;
            }



            int max = int.MinValue;
            for (int i = 0; i < BitsPerSymbol; i++)
            {
                int count = occurrences[i];
                if (count > max)
                    max = count;
            }

            for (int i = 0; i < BitsPerSymbol; i++)
            {
                int count = occurrences[i];
                int len = (int)Math.Round(30.0 * count / (double)max);
                string s1 = String.Format("{0,3:D}", i) + ": " + String.Format("{0,8:D}", count) + " ";

                string s2 = "";
                for (int j = 0; j < len; j++)
                    s2 += "*";

                result.Add(s1 + s2);
            }


            
            
            return result;
        }
    }

    public class SymbolAsBitsAndDistance
    {
        public SymbolAsBits SymbolAsBits;
        public int Distance;

        public override string ToString()
        {
            string result = "";
            for (int i = 0; i < SymbolAsBits.Bits.Length; i++)
                result += SymbolAsBits.Bits[i].ToString();
            result += " ";
            result += SymbolAsBits.Node.ToString();
            return result;
        }
    }


    ///-------------------------------------------------------------------------
    /// <summary>
    /// Provides a bitvector for a symbol.
    /// </summary>
    ///-------------------------------------------------------------------------
    public class SymbolAsBits
    {
        private static Random m_Rnd = new Random();
        private int m_CountOfBits;
        public byte[] Bits;
        public Node Node;

        public SymbolAsBits(Node node, int countOfBits)
        {
            Node = node;
            m_CountOfBits = countOfBits;
            Bits = new byte[countOfBits]; // One byte per bit.

            Init();
        }

        private void Init()
        {
            for (int i = 0; i < m_CountOfBits; i++)
                if (m_Rnd.NextDouble() > 0.5)
                    Bits[i] = 1;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Reduce the distance to another bitvector by making
        /// a randomly chosen bit equal.
        /// </summary>
        ///---------------------------------------------------------------------
        public void AssimilateTo(SymbolAsBits other)
        {
            int idx = m_Rnd.Next(m_CountOfBits);
            this.Bits[idx] = other.Bits[idx];
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Increase the distance of a vector to another one
        /// by flipping one own bit.
        /// </summary>
        ///---------------------------------------------------------------------
        public void IncreaseDistanceTo(SymbolAsBits other)
        {
            int idx = m_Rnd.Next(m_CountOfBits);
            this.Bits[idx] = not(other.Bits[idx]);
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Invert a bit (encoded here as one byte).
        /// </summary>
        ///---------------------------------------------------------------------
        private byte not(byte b)
        {
            if (b == 0)
                return 1;
            else
                return 0;
        }



        public int DistanceTo(SymbolAsBits other)
        {
            int result = 0;
            for (int i = 0; i < this.m_CountOfBits; i++)
            {
                if (this.Bits[i] != other.Bits[i])
                    result++;
            }

            return result;
        }

        public override string ToString()
        {
            string result = "";
            for (int i = 0; i < m_CountOfBits; i++)
                result += Bits[i].ToString();

            result += " " + Node.ToString() ;

            return result;
        }


        public string ToStringOnlyBits()
        {
            string result = "";
            for (int i = 0; i < m_CountOfBits; i++)
                result += Bits[i].ToString();

            return result;
        }





    }
}
