//------------------------------------------------------------------------------
// Purpose: Lisp-like expressions, solely based on symbols and parentheses,
//          in order to support AI applications
//
// Date   : 27. Oct. 2010
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;

namespace Fusym
{
    public class Expression: IDocumentList, IDocument
    {
        public static Encoding encoding = System.Text.Encoding.GetEncoding(1252);



        public static ExpressionList Cons(Expression car, Expression cdr)
        {
            ExpressionList result = new ExpressionList();
            result.Car = car;
            result.Cdr = cdr;
            return result;
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Loads a textfile and returns the content as string.
        /// </summary>
        ///---------------------------------------------------------------------
        public string LoadFromFile(string filename)
        {
            string result = "";

            StreamReader sr = null;

            BufferedStream stream = new BufferedStream(new FileStream(filename, FileMode.Open));

            sr = new StreamReader(stream, encoding);

            StringBuilder sb = new StringBuilder();

            string line;
            while ((line = sr.ReadLine()) != null)
                sb.AppendLine(line.ToUpper());

            sr.Close();

            result = sb.ToString();

            return result;
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Saves a string list as textfile.
        /// </summary>
        ///---------------------------------------------------------------------
        public bool SaveToFile(List<string> sl, string Filename)
        {
            FileStream stream = new FileStream(Filename, FileMode.Create);
            if (!stream.CanWrite)
            {
                stream.Close();
                return false;
            }

            StreamWriter writer = new StreamWriter(stream, encoding);
            foreach (string s in sl)
                writer.WriteLine(s);
            writer.Flush();
            writer.Close();
            stream.Close();
            return true;
        }



        /// <summary>
        /// 
        /// </summary>
        public static IEnumerable<Expression> EnumerateListElements(Expression ex)
        {
            List<string> list = new List<string>();
            if (ex is ExpressionSymbol)
                yield return ex;
            else
            if (ex is ExpressionList)
            {
                Expression car;
                Expression cdr = ex;

                while (cdr is ExpressionList)
                {
                    car = (cdr as ExpressionList).Car;
                    cdr = (cdr as ExpressionList).Cdr;

                    yield return car;
                }


                if (cdr is ExpressionSymbol)
                    yield return cdr;
            }
        }

        ///---------------------------------------------------------------------
        /// <summary>
        /// Creates a sequence of spaces.
        /// </summary>
        ///---------------------------------------------------------------------
        public static string Spaces(int n)
        {
            string result = new string(' ', n);
            return result;
        }

        public static List<string> PrettyPrint(Expression ex, int indentation)
        {
            List<string> result = new List<string>();
            PrettyPrint(ex, indentation, result);
            return result;
        }



        public static List<string> PrettyPrint(Expression ex)
        {
            List<string> result = PrettyPrint(ex, 0);
            return result;
        }
        
        
        private static void PrettyPrint(Expression ex, int indentation, List<string> list)
        {
            string prefix = Spaces(indentation);

            if (ex is ExpressionSymbol)
               list.Add(prefix + (ex as ExpressionSymbol).Symbol);
            else
            if (ex is ExpressionList)
            {

                list.Add(prefix + "(");

                Expression car;
                Expression cdr = ex;

                while (cdr is ExpressionList)
                {
                    car = (cdr as ExpressionList).Car;
                    cdr = (cdr as ExpressionList).Cdr;

                    PrettyPrint(car,indentation + 1, list);
                }


                if (cdr is ExpressionSymbol)
                {
                    list.Add(Spaces(indentation + 1) + ".");
                    PrettyPrint(cdr,indentation + 1, list);
                }


                list.Add(prefix + ")");
            }
        }

        public static List<string> PrettyPrintList(Expression ex)
        {
            List<string> list = new List<string>();
            if (ex is ExpressionSymbol)
                list.Add((ex as ExpressionSymbol).Symbol);
            else
            if (ex is ExpressionList)
            {
                Expression car;
                Expression cdr = ex;

                while (cdr is ExpressionList)
                {
                    car = (cdr as ExpressionList).Car;
                    cdr = (cdr as ExpressionList).Cdr;

                    PrettyPrint(car, 0, list);
                }


                if (cdr is ExpressionSymbol)
                {
                    list.Add(".");
                    PrettyPrint(cdr, 0, list);
                }
            }

            return list;
        }


        public virtual IDocument GetFirstDocument()
        {
            throw new Exception("This method is abstract");
        }

        public virtual IDocument GetNextDocument()
        {
            throw new Exception("This method is abstract.");
        }

        public virtual string GetFirstSymbol()
        {
            throw new Exception("This method is abstract.");
        }

        public virtual string GetNextSymbol()
        {
            throw new Exception("This method is abstract.");
        }

        public virtual ExpressionList FindSymbols()
        {
            throw new Exception("This method is abstract.");
        }

        public static ExpressionList Reverse(ExpressionList list)
        {
            ExpressionList result = null;
            ExpressionList loop = list;
            while (loop != null)
            {
                Expression car = loop.Car;
                loop = loop.Cdr as ExpressionList;
                result = Expression.Cons(car, result);
            }

            return result;
        }
    }

    public class ExpressionSymbol : Expression, IDocumentList, IDocument
    {
        public string Symbol;
        public ExpressionSymbol(string symbol)
        {
            Symbol = symbol;
        }




        public override IDocument GetFirstDocument()
        {
            return this;
        }

        public override IDocument GetNextDocument()
        {
            return null;
        }

        public override string GetFirstSymbol()
        {
            return Symbol;
        }

        public override string GetNextSymbol()
        {
            return "";
        }

    }


    public class ExpressionList : Expression
    {
        public Expression Car = null;
        public Expression Cdr = null;

        private Expression m_CurrentDocument;
        private Expression m_CurrentSymbol;
        private ExpressionList Symbols = null;

        public override IDocument GetFirstDocument()
        {
            m_CurrentDocument = Cdr;
            return Car;
        }

        public override IDocument GetNextDocument()
        {
            IDocument result;


            if (m_CurrentDocument is ExpressionSymbol)
            {
                result = m_CurrentDocument;
                m_CurrentDocument = null;
            }
            else
            if (m_CurrentDocument is ExpressionList)
            {
                result = (m_CurrentDocument as ExpressionList).Car;
                m_CurrentDocument = (m_CurrentDocument as ExpressionList).Cdr;
            }
            else
                return null;

            return result;
        }

        public override string GetFirstSymbol()
        {
            Symbols = FindSymbols();
            m_CurrentSymbol = Symbols;
            return GetNextSymbol();
        }

        public override string GetNextSymbol()
        {
            if (m_CurrentSymbol == null)
                return "";

            string result = (((m_CurrentSymbol as ExpressionList).Car) as ExpressionSymbol).Symbol;
            m_CurrentSymbol = (m_CurrentSymbol as ExpressionList).Cdr;
            return result;
        }


        public override ExpressionList FindSymbols()
        {
            ExpressionList result;

            if (Cdr != null)
                result = Cdr.FindSymbols();
            else
                result = null;

            if (Car is ExpressionSymbol)
                result = Cons(Car, result);
            else
            if (Car is ExpressionList)
            {
                ExpressionList loop = Car.FindSymbols();
                while (loop != null)
                {
                    result = Cons(loop.Car,result);
                    loop = loop.Cdr as ExpressionList;
                }
            }

            return result;
        }
    }








    public class Tokenizer
    {
        private string s;
        Regex regex = new Regex(@"(?<SPACE>\s+)|" +
                                @"(?<OPENPARENTHESE>\()|" +
                                @"(?<CLOSEPARENTHESE>\))|" +
                                @"(?<SYMBOL>[^\(\)\s]+)" // Not '(' and not ')' and no space.
                               );


        Match m;

        public string TokenType;
        public string TokenValue;

        public Tokenizer(string toTokenize)
        {
            s = toTokenize;
        }

        public void FirstToken()
        {
            m = regex.Match(s);
            SetToken();
        }

        public void NextToken()
        {
            m = m.NextMatch();
            SetToken();
        }

        public void FirstTokenSkippingSpaces()
        {
            FirstToken();
            while (TokenType.Equals("SPACE"))
                NextToken();
        }


        public void NextTokenSkippingSpaces()
        {
            NextToken();
            while (TokenType.Equals("SPACE"))
                NextToken();
        }


        public void SetToken()
        {
            if (m.Success)
            {
                TokenType  = "UNEXPECTED";
                TokenValue = "UNEXPECTED";

                for (int i = 1; i < m.Groups.Count; i++)
                {
                    int countOfFoundMatches = 0;

                    Group g = m.Groups[i];
                    CaptureCollection cc = g.Captures;
                    for (int j = 0; j < cc.Count; j++)
                    {
                        countOfFoundMatches++;
                        Capture c = cc[j];
                        TokenType = regex.GroupNameFromNumber(i); // Name of the current group
                        TokenValue = c.ToString();
                        // Es gibt noch c.Index, das die Position angibt. Evtl. fuer Fehlermeldungen
                    }

                    if (cc.Count > 1)
                        throw new Exception(String.Format("Tokenizer.SetToken: illegal count of matches {0}", countOfFoundMatches));
                }


            }
            else
            {
                TokenType  = "EOF";
                TokenValue = "EOF";
            }
        }

    }


    public class Parser
    {
        private Tokenizer tokenizer;
        private string TokenType;
        private string TokenValue;

        public Parser(string toParse)
        {
            tokenizer = new Tokenizer(toParse);
        }

        private void consume(string tokentype)
        {
            if (TokenType != tokentype)
                throw new Exception(String.Format("Syntax error: expected {0} but got '{1}'", tokentype, TokenValue));

            tokenizer.NextTokenSkippingSpaces();
            TokenType = tokenizer.TokenType;
            TokenValue = tokenizer.TokenValue;
        }

        public Expression Parse()
        {
            Expression result;

            tokenizer.FirstTokenSkippingSpaces();
            TokenType  = tokenizer.TokenType;
            TokenValue = tokenizer.TokenValue;


            result = ImplicitParentheses();
            return result;
        }


        ExpressionList ImplicitParentheses()
        {
            ExpressionList result = null;

            while (tokenizer.TokenType != "EOF")
            {
                Expression e = Factor();
                result = ExpressionList.Cons(e, result);
            }

            consume("EOF");
            return Expression.Reverse(result);
        }


        Expression Factor()
        {
            Expression result;

            if (TokenType.Equals("OPENPARENTHESE"))
            {
                ExpressionList list = null;
                consume("OPENPARENTHESE");
                while (tokenizer.TokenType != "CLOSEPARENTHESE")
                {
                    Expression e = Factor();
                    list = Expression.Cons(e, list);
                }

                consume("CLOSEPARENTHESE");
                result = Expression.Reverse(list);
            }
            else
            {
                result = new ExpressionSymbol(TokenValue.ToUpper());
                consume("SYMBOL");
            }

            return result;
        }

    }

}
