//------------------------------------------------------------------------------
// Purpose: Display a double number distribution as a spectrum
// Date   : 30.Jul.2008
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace Unarch5
{
    public interface IDistributionProvider
    {
     double GetFirstDistributionValue();
     double GetNextDistributionValue();
    }

    class DistributionBar
    {
        private IDistributionProvider m_DistributionProvider;
        private List<double> m_Fractions = new List<double>();
        private Control m_Control;
        private int m_Left;
        private int m_Top;
        private int m_Height;
        private int m_Width;
        private int m_CountOfLines;
        private int m_CountOfDistributionValues;
        private double m_Min;
        private double m_Max;
                               
        public DistributionBar(
                               IDistributionProvider distributionProvider,
                               Control control
                              )
        {
            m_DistributionProvider = distributionProvider;        
            m_Control = control;
        }
        

        
        private void Prepare()
        {
            double d;
            
            m_Top    = 1;
            m_Left   = 1;
            m_Height = m_Control.Height - 2;
            m_Width  = m_Control.Width - 2;
            m_CountOfLines = m_Width;

            m_Min = -1.0;
            m_Max = -1.0;
            m_CountOfDistributionValues = 0;
            
            d = m_DistributionProvider.GetFirstDistributionValue();
            while (d >= 0.0)
            {
                d = m_DistributionProvider.GetNextDistributionValue();
                
                if ((d < m_Min) || (m_Min < 0))
                   m_Min = d;
                   
                if ((d > m_Max) || (m_Max < 0))
                   m_Max = d;
                   
                m_CountOfDistributionValues++;
                if (m_CountOfDistributionValues >= m_CountOfLines)
                   break;                
            }

            m_Fractions.Clear();            
            m_Fractions.Capacity = m_CountOfDistributionValues;
            
            
            
            int i = 0;
            d = m_DistributionProvider.GetFirstDistributionValue();
            while (d >= 0.0)
            {
                d = m_DistributionProvider.GetNextDistributionValue();
                if (m_Max > m_Min)
                   m_Fractions.Add((d-m_Min)/(m_Max-m_Min));
                else
                   m_Fractions.Add(1.0);   
                   
                   
                i++;
                if (i >= m_CountOfLines)
                   break;                
            }
        }
        

        ///---------------------------------------------------------------------
        /// <summary>
        /// 0.0 <= fraction <= 1.0
        /// </summary>
        /// <param name="fraction"></param>
        /// <param name="r"></param>
        /// <param name="g"></param>
        /// <param name="b"></param>
        ///---------------------------------------------------------------------
        private void GetRGB(double fraction, out int r, out int g, out int b)
        {
            if (fraction < 0.25)
            {
                fraction *= 4.0;
                r = 255;
                g = (int)Math.Round(255.0 * fraction);
                b = 0;
            }
            else
            if (fraction < 0.5)
            {
                fraction -= 0.25;
                fraction *= 4.0;
                r = 255 - (int)Math.Round(255.0 * fraction);
                g = 255;
                b = 0;
            }
            else
            if (fraction < 0.75)
            {
                fraction -= 0.5;
                fraction *= 4.0;
                r = 0;
                g = 255;
                b = (int)Math.Round(255.0 * fraction);
            }
            else
            {
                fraction -= 0.75;
                fraction *= 4.0;
                r = 0;
                g = 255 - (int)Math.Round(255.0 * fraction);
                b = 255;
            }
        }
    
    
        public void Paint(PaintEventArgs e)
        {
            Graphics graphics = e.Graphics;
           
            Prepare();
            for (int i = 0; i < m_CountOfLines; i++)
            {
                double fraction;
                
                if (m_CountOfDistributionValues > 0)
                {
                   double pos;
                   
                   pos = i / (double)m_CountOfLines;
//                   pos = pos * pos * pos * pos;
                   
                   
                   int index = (int)Math.Round(m_CountOfDistributionValues * pos);
                   if (index >= m_CountOfDistributionValues)
                      index = m_CountOfDistributionValues - 1;
                   fraction = m_Fractions[index];
                }
                else
                   fraction = 0.0;                   
                   
                int r,g,b;
                GetRGB(1.0-fraction, out r, out g, out b);
                  
                Pen pen = new Pen(Color.FromArgb(r,g,b),1);
                int x1 = m_Left+i;
                int y1 = m_Top + (int)Math.Round(0.5 * (m_Height - m_Height * fraction));
                int x2 = m_Left+i;
                int y2 = m_Top + (int)Math.Round(0.5 * (m_Height + m_Height * fraction));
                
                
                
                graphics.DrawLine(pen,x1,y1,x2,y2);
            }
        }    
    }
}
