//------------------------------------------------------------------------------
// Purpose: Display a semantic network
// Date   : 7.Feb.2007
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using Fusym;
using System.Text.RegularExpressions;

namespace Unarch5
{
    public partial class FormGraph : Form
    {
        Hashtable symbolToLabel = new Hashtable();
        Random rndObj  = new Random();
        SemanticNetwork cm; 

        public FormGraph()
        {
            InitializeComponent();

            cm = GlobalClusterManager.GetInstance().SemanticNetwork;
        }


        private void LoadSettings()
        {
            Configuration c = Configuration.GetInstance();
            this.Location = c.FormGraphSettings.Location;
            this.Size     = c.FormGraphSettings.Size;
        }


        private void StoreSettings()
        {
            Configuration c = Configuration.GetInstance();
            c.FormGraphSettings.Location = this.Location;
            c.FormGraphSettings.Size     = this.Size;
        }




        private void FormGraph_DragEnter(object sender, DragEventArgs e)
        {
            // If the data is a file or a bitmap, display the copy cursor.
            if (e.Data.GetDataPresent(typeof(MyLabel)))
            {
                e.Effect = DragDropEffects.Copy;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }

        }


        private void label1_MouseDown(object sender, MouseEventArgs e)
        {
            if (!radioButtonMove.Checked)
                return;

            DataObject data = new DataObject(sender);
            DoDragDrop(data, DragDropEffects.All);
        }


        /// <summary>
        /// Create a MyLabel instance, if not already there.
        /// </summary>
        /// <param name="symbol"></param>
        private void AddSymbol(string symbol, bool bold)
        {
            if (! symbolToLabel.ContainsKey(symbol))
               {
                MyLabel lbl = new MyLabel();
                lbl.SuspendLayout();
                lbl.Parent = this;
                lbl.Text = symbol;
                
                if (bold)
                   lbl.Font = new Font(lbl.Font, FontStyle.Bold); 

                lbl.AutoSize = true;
                lbl.BackColor = Color.White;
                lbl.Left = rndObj.Next(Width);
                lbl.Top  = panelTop.Height + rndObj.Next(Height-panelTop.Height);
                lbl.MouseDown   += label1_MouseDown;
                lbl.DoubleClick += label1_DoubleClick;
                lbl.Paint += label1_Paint;
                lbl.ResumeLayout();
                symbolToLabel[symbol] = lbl;

                AddNeighbourLabels(symbol,lbl);
            }
        }
        
        private void AddSymbol(string symbol)
        {
            AddSymbol(symbol,false);
        }
        
        

        private void AddNeighbourLabels(string symbol, MyLabel lbl)
        {
            Node n = cm.GetNode(symbol);
            if (n == null)
                return;

            Node m = n.GetFirstNeighbour();
            while (m != null)
            {
                string s = m.ToString();
                MyLabel l = (MyLabel)symbolToLabel[s];
                if (l != null)
                   lbl.AddNeighbour(l);

                m = n.GetNextNeighbour();
            }
        }


        private void buttonAdd_Click(object sender, EventArgs e)
        {
            string [] symbols = Regex.Split(textBoxAdd.Text.ToUpper(),@"\s+");

            foreach (string symbol in symbols)
                AddSymbol(symbol,true);

            
            foreach (string symbol1 in symbols)
            {
                foreach (string symbol2 in symbols)
                    if (symbol1.CompareTo(symbol2) < 0)
                        {
                            List<Node> path = new List<Node>();
                            cm.FindPath(symbol1,symbol2,path);
                            
                            foreach (Node n in path)
                            {
                                string symbol3 = n.ToString();
                                AddSymbol(symbol3);
                            }
                        }
            }
                
            Invalidate();
        }


        private void FormGraph_DragDrop(object sender, DragEventArgs e)
        {
            Point clientPoint = this.PointToClient(new Point(e.X, e.Y));

            MyLabel l = (MyLabel)e.Data.GetData(typeof(MyLabel));
            l.Left = clientPoint.X - l.Width / 2;
            l.Top  = clientPoint.Y;

            Invalidate();
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            Close();
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// If a MyLabel instance is doubleclicked, add its symbols m_Neighbours.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        ///---------------------------------------------------------------------
        private void label1_DoubleClick(object sender, EventArgs e)
        {
            if (!radioButtonExpand.Checked)
                return;


            Cursor.Current = Cursors.WaitCursor;

            string symbol = (sender as MyLabel).Text;

            Node n = cm.GetNode(symbol);
            if (n == null)
                return;

            Node m = n.GetFirstNeighbour();
            while (m != null)
            {
                AddSymbol(m.ToString());
                m = n.GetNextNeighbour();
            }

            Cursor.Current = Cursors.Default;
            Invalidate();
        }

        private void textBoxAdd_Enter(object sender, EventArgs e)
        {
            this.AcceptButton = buttonAdd;
        }


        private void FormGraph_Paint(object sender, PaintEventArgs e)
        {
            PaintConnectionsBetweenLabels(e);
        }


        private void PaintConnectionsBetweenLabels(PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            Pen pen = new Pen(Color.Red, 1);


            foreach (Control c in Controls)
            {
                if (c is MyLabel)
                {
                    int x1 = (c as MyLabel).Left + (c as MyLabel).Width / 2;
                    int y1 = (c as MyLabel).Top  + (c as MyLabel).Height / 2;

                    List<MyLabel> list = (c as MyLabel).NeighbourLabels;
                    foreach (MyLabel m in list)
                    {
                     //   if ((c as MyLabel).GetHashCode() < m.GetHashCode())
                     //       continue;


                        int x2 = m.Left + m.Width / 2;
                        int y2 = m.Top  + m.Height / 2;

                        g.DrawLine(pen,
                                   new Point(x1, y1),
                                   new Point(x2, y2));
                    }

                    //(c as MyLabel).Invalidate();
                }

            }
        }








        private void buttonClear_Click(object sender, EventArgs e)
        {
            for (int i = Controls.Count - 1; i >= 0; i--)
            {
                if (Controls[i] is MyLabel)
                    Controls.RemoveAt(i);
            }

            symbolToLabel.Clear();
            Invalidate();
        }


        private void label1_Paint(object sender, PaintEventArgs e)
        {
            if (! (sender is MyLabel))
               return;

            MyLabel c = (MyLabel)sender;


            Graphics g = e.Graphics;
            Pen pen = new Pen(Color.Red, 1);

            int x1 = c.Left + c.Width  / 2;
            int y1 = c.Top  + c.Height / 2;

            List<MyLabel> list = c.NeighbourLabels;
            foreach (MyLabel m in list)
            {
                if (c.GetHashCode() < m.GetHashCode())
                    continue;


                int x2 = m.Left + m.Width  / 2;
                int y2 = m.Top  + m.Height / 2;

                g.DrawLine(pen,
                           new Point(x1, y1),
                           new Point(x2, y2));
            }

        }



    }

    internal class MyLabel : Label
    {
        private List<MyLabel> m_NeighbourLabels = new List<MyLabel>();


        public List<MyLabel> NeighbourLabels
        {
            get { return m_NeighbourLabels; }
        }
	

        public void AddNeighbour(MyLabel l)
        {
            m_NeighbourLabels.Add(l);
        }

        public void Clear()
        {
            m_NeighbourLabels.Clear();
        }
    }

}
