//------------------------------------------------------------------------------
// Purpose: Search form. Allows the user to put in some words and to
//          search for documents.
//
// Date   : 22.Jun.2006
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.Text.RegularExpressions;
using System.IO;
using Fusym;
using System.Threading;

namespace Unarch5
{
    public partial class FormSearchAssociative : Form, ISearchProgressReceiver
    {
        private string           m_Filename;
        private IDocumentList    m_DocumentList;
        private IClusterProvider m_AdditionalClusters;
        private ValuedDocList    m_ValuedDocList;

        public string Query
        {
            get
            {
                List<string> inputSymbols = StringUtils.GetQuerySymbols(cbQuery.Text);
                string result = "";

                foreach (string s in inputSymbols)
                    result += s + " ";

                return result.TrimEnd();
            }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="filename">The document on which the search is applied.</param>
        /// <param name="documentlist"></param>
        /// <param name="additionalClusters"></param>
        /// <param name="valuedDocList"></param>
        ///---------------------------------------------------------------------
        public FormSearchAssociative(
                                     string           filename, 
                                     IDocumentList    documentlist, 
                                     IClusterProvider additionalClusters, 
                                     ValuedDocList    valuedDocList
                                    )
        {
            InitializeComponent();

            m_Filename           = filename;
            m_DocumentList       = documentlist;
            m_AdditionalClusters = additionalClusters;
            m_ValuedDocList      = valuedDocList;
        }

        private void FormSearchAssociative_Load(object sender, EventArgs e)
        {
            numMaxDepth.Value = 3;
            tbMaxSym.Text = "10000";
            LoadSettings();
            ActiveControl = cbQuery;

            checkBoxSearchSymbols_CheckStateChanged(null, null);
        }





        ///---------------------------------------------------------------------
        /// <summary>
        /// Load all cluster files that are mentioned in the configuration 
        /// settings as active cluster files.
        /// </summary>
        ///---------------------------------------------------------------------
        private void LoadClusterFiles(SemanticNetwork sn)
        {
            Configuration c = Configuration.GetInstance();
            List<string> activeFiles = c.ActiveClusterFiles;

            foreach (string filename in activeFiles)
            {
                //statusStrip.Text = filename;
                toolStripStatusLabel.Text = filename;
                Application.DoEvents();
                try
                {
                    sn.LoadFromFile(filename);
                }
                catch
                {
                }
            }

            toolStripStatusLabel.Text = "";
            Application.DoEvents();
        }













        private void FillQueryText(ItemCheckEventArgs e)
        {
            cbQuery.Text = "";

            for (int i = 0; i < checkedListBoxSearchSymbols.Items.Count; i++)
            {
                if (
                    (
                     checkedListBoxSearchSymbols.GetItemChecked(i) &&
                     (((e.Index != i)) || (e.NewValue == CheckState.Checked)) // is not going to be unchecked
                    )
                    
                    ||

                    ((e.Index == i) && (e.NewValue == CheckState.Checked)) // is going to be checked
                   ) 
                   cbQuery.Text += checkedListBoxSearchSymbols.Items[i] + " ";
            }
        }







        ///---------------------------------------------------------------------
        /// <summary>
        /// Search for other symbols.
        /// </summary>
        ///---------------------------------------------------------------------
        private void FillCheckboxes(IDocumentEvaluator de)
        {
            List<string> inputSymbols = StringUtils.GetQuerySymbols(cbQuery.Text);
            List<string> associatedSymbols = de.AssociatedSymbols;

            
            
            ValuedDocList vdl = new ValuedDocList();

            foreach (string s in associatedSymbols)
            {
                OneStringDocument doc = new OneStringDocument(s);
                double value = de.Evaluate(doc);
                vdl.Add(doc, value);
            }


            checkedListBoxSearchSymbols.BeginUpdate();
            checkedListBoxSearchSymbols.Items.Clear();
            for (int i = 0; i < vdl.Count; i++)
            {
                ValuedDocument vdoc = vdl[i];
                checkedListBoxSearchSymbols.Items.Add(((OneStringDocument)vdoc.Object).Text);
            }

            //--- Set all strings of the query checked ---
            foreach (string s in inputSymbols)
            {
                int idx = checkedListBoxSearchSymbols.FindStringExact(s);
                if (idx >= 0)
                   checkedListBoxSearchSymbols.SetItemChecked(idx, true);
            }
            checkedListBoxSearchSymbols.EndUpdate();
        }






















        private void buttonSearch_Click(object sender, EventArgs e)
        {
            Search();  
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Perform the search.
        /// </summary>
        ///---------------------------------------------------------------------
        private void Search()
        {
            Cursor.Current = Cursors.WaitCursor;


            cbQuery.Text = cbQuery.Text.ToUpper();
            if ((cbQuery.Items.IndexOf(cbQuery.Text) < 0) && (cbQuery.Text.Length > 0))
                cbQuery.Items.Insert(0, cbQuery.Text); // History

            try
            {
                int maxSym = 10000;
                int maxDepth = 5;


                try
                {
                    maxSym = Int32.Parse(tbMaxSym.Text);
                }
                catch
                {
                    // Keep the default
                }



                try
                {
                    maxDepth = (int)numMaxDepth.Value;
                }
                catch
                {
                    // Keep the default
                }

                //--- Semantic network -----------------------------------------
                SemanticNetwork semanticNetwork = new SemanticNetwork(Configuration.encoding);

                // Load from the calling document
                toolStripStatusLabel.Text = m_Filename;
                Application.DoEvents();


                // Maybe the current file contains associations
                if (checkBoxClusterProvider.Checked)
                    semanticNetwork.LoadFromClusterProvider(m_AdditionalClusters);


                LoadClusterFiles(semanticNetwork);
                //--- DocumentEvaluator ----------------------------------------


                List<string> inputSymbols = StringUtils.GetQuerySymbols(cbQuery.Text);

                EvaluationMethod evm = (EvaluationMethod)cbMethod.SelectedIndex;

                IDocumentEvaluator documentEvaluator;

                if (evm == EvaluationMethod.RandomWalk)
                    documentEvaluator = new DocumentEvaluatorRandomWalk(semanticNetwork, inputSymbols, maxDepth, maxSym);
                else
                    documentEvaluator = semanticNetwork.NewDocumentEvaluator(
                                                                              inputSymbols,
                                                                              maxDepth,
                                                                              maxSym,
                                                                              evm,
                                                                              checkBoxSubStringSearch.Checked
                                                                             );


                if (checkBoxSearchSymbols.Checked)
                {
                    FillCheckboxes(documentEvaluator);
                    return;
                }

                //--------------------------------------------------------------


                SearchManager sm = new SearchManager(
                                                     this,
                                                     m_DocumentList, 
                                                     m_ValuedDocList, 
                                                     documentEvaluator,
                                                     semanticNetwork
                                                    );


                sm.Run();

                if (!checkBoxSearchSymbols.Checked)
                    Close();
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }

        }


        void ISearchProgressReceiver.SetCountOfDocuments(int count)
        {
            toolStripProgressBar.Maximum = count;
        }

        void ISearchProgressReceiver.SetIndexOfDocument(int idx)
        {
            if (idx % 100 == 0)
               toolStripProgressBar.Value = idx;
        }



        private void buttonClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void FormSearchAssociative_FormClosed(object sender, FormClosedEventArgs e)
        {
            StoreSettings();
        }



        private void LoadSettings()
        {
            Configuration c = Configuration.GetInstance();
            this.Location               = c.FormSearchAssociativeSettings.Location;
            this.numMaxDepth.Value      = c.FormSearchAssociativeSettings.MaxDepth;
            this.tbMaxSym.Text          = c.FormSearchAssociativeSettings.MaxSym.ToString();
            this.cbMethod.SelectedIndex = c.FormSearchAssociativeSettings.EvaluationMethod;
            this.checkBoxClusterProvider.Checked = c.FormSearchAssociativeSettings.FileYieldsAssociations;
            this.cbQuery.Items.Clear();
            this.cbQuery.Items.AddRange(c.History.ToArray());
            this.checkBoxSubStringSearch.Checked = c.FormSearchAssociativeSettings.FullTextSearch;
        }




        private void StoreSettings()
        {
            int maxSym = 100000;
            try
            {
                maxSym = Int32.Parse(tbMaxSym.Text);
            }
            catch
            {
                // Keep the default
            }


            Configuration c = Configuration.GetInstance();
            c.FormSearchAssociativeSettings.Location = this.Location;
            c.FormSearchAssociativeSettings.MaxDepth = (int)this.numMaxDepth.Value;
            c.FormSearchAssociativeSettings.MaxSym   = maxSym;
            c.FormSearchAssociativeSettings.EvaluationMethod = this.cbMethod.SelectedIndex;
            c.FormSearchAssociativeSettings.FileYieldsAssociations = this.checkBoxClusterProvider.Checked;
            c.FormSearchAssociativeSettings.FullTextSearch = this.checkBoxSubStringSearch.Checked;
            c.History.Clear();
            AddToStringList(c.History, this.cbQuery.Items);
            c.RestrictArrayListLength(c.History, c.MaxQueryHistory);
          }



        private void AddToStringList(List<string> list, ComboBox.ObjectCollection objects)
        {
            foreach (string s in objects)
                list.Add(s);
        }


        private void cbQuery_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (Char.IsLetter(e.KeyChar))
               e.KeyChar = Char.ToUpper(e.KeyChar);
        }

        private void FormSearchAssociative_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)27)
               Close(); 
        }

        private void checkBoxSearchSymbols_CheckStateChanged(object sender, EventArgs e)
        {
            if (checkBoxSearchSymbols.Checked)
               {
                   this.Height = 400;
                   this.FormBorderStyle = FormBorderStyle.Sizable;
                   checkedListBoxSearchSymbols.Visible = true;
               }
            else
               {
                   this.Height = 230;
                   this.FormBorderStyle = FormBorderStyle.FixedDialog;
                   checkedListBoxSearchSymbols.Visible = false;
               }
        }



        private void checkedListBoxSearchSymbols_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            FillQueryText(e);
        }
    }
}