//------------------------------------------------------------------------------
// Purpose: Editor for Unarch files.
// Date   : 14.Jun.2006
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Fusym;

namespace Unarch5
{
    public partial class FormStandardEditor : Form
    {
        DataSetState   m_State;
        LineRecordFile lrf = new LineRecordFile(Configuration.encoding);
        ValuedDocList m_ValuedDocList = new ValuedDocList(); 
        int current = 0;

        string m_Filename;


        private enum DataSetState
        {
            dsBrowsing,
            dsEditing,
            dsInsert
        }

        public FormStandardEditor()
        {
            InitializeComponent();
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Actions when state changes.
        /// </summary>
        /// <param name="state"></param>
        ///---------------------------------------------------------------------
        private void SetState(DataSetState state)
        {
            m_State = state;
            if (m_State == DataSetState.dsBrowsing)
               {
                   buttonFirst.Enabled = true;
                   buttonPrev.Enabled = true;
                   buttonNext.Enabled = true;
                   buttonLast.Enabled = true;
                   buttonEdit.Enabled = true;
                   buttonInsert.Enabled = true;
                   buttonDelete.Enabled = true;
                   buttonStore.Enabled = false;
                   buttonCancel.Enabled = false;
                   textBoxInfo.ReadOnly = true;
                   textBoxInfo.BackColor = Color.White;
               }
            else
                   if ((m_State == DataSetState.dsEditing) || (m_State == DataSetState.dsInsert))
               {
                   buttonFirst.Enabled = false;
                   buttonPrev.Enabled = false;
                   buttonNext.Enabled = false;
                   buttonLast.Enabled = false;
                   buttonEdit.Enabled = false;
                   buttonInsert.Enabled = false;
                   buttonDelete.Enabled = false;
                   buttonStore.Enabled = true;
                   buttonCancel.Enabled = true;
                   textBoxInfo.ReadOnly = false;
               }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Load the record that corresponds to the current index
        /// </summary>
        ///---------------------------------------------------------------------
        private void ApplyCurrent()
        {
            if ((current >= 0) && (current < m_ValuedDocList.Count))
            {
                ValuedDocument vdoc = m_ValuedDocList[current];
                LineRecord lrec = (LineRecord)vdoc.Object;
                labelValue.Text      = vdoc.Value.ToString();
                textBoxInfo.Text = lrec.Text;
            }
            else
                textBoxInfo.Clear();
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Make sure that all records of lrf are in m_ValuedDocList. Value = index
        /// </summary>
        ///---------------------------------------------------------------------
        private void SelectAll()
        {
            
            m_ValuedDocList.Clear();
            for (int i = 0; i < lrf.Count; i++)
            {
                double value = i;
                m_ValuedDocList.Add(lrf.Records[i],value);
            }

            current = 0;
            ApplyCurrent();
        }


        private void buttonFirst_Click(object sender, EventArgs e)
        {
            current = 0;
            ApplyCurrent();
        }

        private void buttonPrev_Click(object sender, EventArgs e)
        {
            current--;
            if (current < 0)
                current = 0;
            ApplyCurrent();
        }

        private void buttonNext_Click(object sender, EventArgs e)
        {
            current++;
            if (current >= m_ValuedDocList.Count)
                current = m_ValuedDocList.Count - 1;
            ApplyCurrent();
        }

        private void buttonLast_Click(object sender, EventArgs e)
        {
            current = m_ValuedDocList.Count - 1;
            ApplyCurrent();
        }

        private void LoadFile(string filename)
        {
            Cursor.Current = Cursors.WaitCursor;
            lrf.LoadFromFile(filename);
            m_Filename = filename;
            SelectAll();
            SetState(DataSetState.dsBrowsing);
            Cursor.Current = Cursors.Default;
        }



        private void buttonEdit_Click(object sender, EventArgs e)
        {
            SetState(DataSetState.dsEditing);
        }


        private void buttonDelete_Click(object sender, EventArgs e)
        {
            if ((current >= 0) && (current < m_ValuedDocList.Count))
            {
                ValuedDocument vdoc = m_ValuedDocList[current];
                LineRecord lrec = (LineRecord)vdoc.Object;

                lrec.Delete();
                textBoxInfo.Clear();
            }
        }

        private void buttonInsert_Click(object sender, EventArgs e)
        {
            SetState(DataSetState.dsInsert);
            textBoxInfo.Clear();
            ActiveControl = textBoxInfo;
        }



        private void buttonStore_Click(object sender, EventArgs e)
        {
            if (m_State == DataSetState.dsEditing)
               {
                   ValuedDocument vdoc = m_ValuedDocList[current];
                   LineRecord lrec = (LineRecord)vdoc.Object;
                   lrec.Text = textBoxInfo.Text;
               }
            else
            if (m_State == DataSetState.dsInsert)
               {
                LineRecord lrec = new LineRecord();
                lrec.Text = textBoxInfo.Text;
                lrf.Add(lrec);
                SelectAll();
               }

            SetState(DataSetState.dsBrowsing);
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            SetState(DataSetState.dsBrowsing);
            //--- Restore old content ---
            ApplyCurrent();

        }


        private void Search(string searched)
        {
            searched = searched.ToUpper().Trim();

            m_ValuedDocList.Clear();
            for (int i = 0; i < lrf.Count; i++)
            {
                LineRecord lrec = lrf.Records[i];

                string s = lrec.GetFirstSymbol().ToUpper();
                while (s != "")
                {
                    if (s == searched)
                    {
                        m_ValuedDocList.Add(lrec, searched,i);
                        s = "";
                    }
                    else
                        s = lrec.GetNextSymbol().ToUpper();
                }
            }

            current = 0;
            ApplyCurrent();
        }

        


        private void buttonSearchExact_Click(object sender, EventArgs e)
        {
            using (FormSearchExact f = new FormSearchExact())
            {
                f.ShowDialog();


                if (f.Result == true)
                {
                    Cursor.Current = Cursors.WaitCursor;
                    Search(f.SearchString);
                    Cursor.Current = Cursors.Default;
                }
            }
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            lrf.SaveToFile(m_Filename); // Das reicht evtl nicht zu-tun
            Close();
        }

        private void buttonSelectAll_Click(object sender, EventArgs e)
        {
            SelectAll();
        }


        private void FormStandardEditor_Load(object sender, EventArgs e)
        {
            LoadSettings();
        }


        private void FormStandardEditor_FormClosed(object sender, FormClosedEventArgs e)
        {
            StoreSettings();
        }


        private void LoadSettings()
        {
            Configuration c = Configuration.GetInstance();
            this.Location = c.FormStandardEditorSettings.Location;
            this.Size     = c.FormStandardEditorSettings.Size;
        }


        private void StoreSettings()
        {
            Configuration c = Configuration.GetInstance();
            c.FormStandardEditorSettings.Location = this.Location;
            c.FormStandardEditorSettings.Size = this.Size;
        }

        private void buttonBrowseFile_Click(object sender, EventArgs e)
        {
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                LoadFile(openFileDialog.FileName);
            }
        }

        private void buttonSearchAssociative_Click(object sender, EventArgs e)
        {
            using (FormSearchAssociative f = new FormSearchAssociative(m_Filename, lrf, lrf, m_ValuedDocList))
            {
                f.ShowDialog();
                current = 0;
                ApplyCurrent();
            }
        }
    }
}