//------------------------------------------------------------------------------
// Purpose: Display statistics.
// Date   : 02. Jan. 2012
// Author : Joachim Pimiskern
// Status : Freeware
//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Fusym;
using System.Threading;
using System.IO;

namespace Unarch5
{
    public partial class FormStatistics : Form
    {
        private Thread m_ThreadDegrees;
        private object m_lockObject = new object();
        private int[] m_DegreeArray = null;
        private int m_MaxDegree = 0;
        private int m_Alpha = 0;


        public FormStatistics()
        {
            InitializeComponent();
            StartThreads();
            WaitForThreadEnd();
            Display();
        }

        private void Display()
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i <= m_MaxDegree; i++)
                sb.AppendLine(i.ToString() + " " + m_DegreeArray[i].ToString());
            textBox1.Text = sb.ToString();


            if (m_MaxDegree != 0)
            {
                double inv = 100.0 / (double)m_Alpha;

                sb = new StringBuilder();
                for (int i = 0; i <= m_MaxDegree; i++)
                {
                    int len = (int)Math.Round((double)m_DegreeArray[i] * inv);
                    string s = String.Format("{0,6:D}", i) + ' ' + new string('*', len);
                    sb.AppendLine(s);
                }

                textBox2.Text = sb.ToString();
            }
        }

        private void StartThreads()
        {
            m_ThreadDegrees = new Thread(ExecuteDegreesThread);
            m_ThreadDegrees.Name = "DegreesThread";
            m_ThreadDegrees.Priority = ThreadPriority.BelowNormal;
            m_ThreadDegrees.Start();
        }

        private void WaitForThreadEnd()
        {
            while (m_ThreadDegrees.IsAlive)
            {
                Application.DoEvents();
            }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Determine degree distribution.
        /// </summary>
        ///---------------------------------------------------------------------
        void ExecuteDegreesThread()
        {
            GlobalClusterManager gm = GlobalClusterManager.GetInstance();
            SemanticNetwork sn = gm.SemanticNetwork;


            // Determine actual maximum count of neighbors
            int maxDeg = 0;
            int count = sn.NodeList.Count;
            for (int i = 0; i < count; i++)
            {
                Node n = sn.NodeList[i];
                int countNeighbors = n.Neighbours.Count;
                if (countNeighbors > maxDeg)
                    maxDeg = countNeighbors;
            }


            // For each count of neighbors, determine how
            // many nodes have that neighborcount.
            int[] array = new int[maxDeg + 1];
            for (int i = 0; i < count; i++)
            {
                Node n = sn.NodeList[i];
                int countNeighbors = n.Neighbours.Count;
                array[countNeighbors]++;
            }

            int alpha = 0;
            for (int i = 0; i <= maxDeg; i++)
            {
                int a = array[i];
                if (a > alpha)
                    alpha = a;
            }

            lock (m_lockObject)
            {
                m_DegreeArray = array;
                m_MaxDegree = maxDeg;
                m_Alpha = alpha;
            }
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void buttonExport_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
            saveFileDialog.Filter = "Textdateien (*.txt)|*.txt|Alle Dateien (*.*)|*.*";
            
            if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
                return;




            List<string> sl = new List<string>();
            string s = "t\tx";
            sl.Add(s);

            for (int i = 0; i <= m_MaxDegree; i++)
            {
                s = i.ToString() + "\t" + m_DegreeArray[i].ToString();
                sl.Add(s);
            }

            File.WriteAllLines(saveFileDialog.FileName, sl.ToArray());
        }
    }
}