// Joachim Pimiskern, 6.12.2009 (?)
using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using Fusym;


namespace Unarch5
{
    ///-------------------------------------------------------------------------
    /// <summary>
    /// The DataSetState is important for the GUI. 
    /// </summary>
    ///-------------------------------------------------------------------------
    public enum DataSetState
    {
        dsBrowsing,
        dsEditing,
        dsInsert
    }



    ///-------------------------------------------------------------------------
    /// <summary>
    /// A LineRecord is a coherent sequence of nonempty lines.
    /// </summary>
    ///-------------------------------------------------------------------------
    public class LineRecord : IDocument
    {
        public List<string> Lines;

        int       m_EnumerationCurrentLine;
        int       m_EnumerationListIndex;
        List<string> m_EnumerationList = new List<string>();

        protected static Regex  m_EnumerationRegex  = new Regex(@"[^\s]+", RegexOptions.Compiled);
        public static string m_WordlistSplitRegex = @"(?:[;,`'""()\n\s])";
        public bool Deleted = false;

        public LineRecord(int capacity)
        {
            Lines = new List<string>(capacity);
        }

        public LineRecord()
        {
            Lines = new List<string>();
        }

        public virtual void Delete()
        {
            Deleted = true;
            Lines.Clear();
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Enumeration of all symbols of one record.
        /// </summary>
        /// <returns>The first symbol, or "" as end marker</returns>
        ///---------------------------------------------------------------------
        public virtual string GetFirstSymbol()
        {
            m_EnumerationCurrentLine = 0;
            return GetNextSymbol();
        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Enumeration of all symbols of one record.
        /// </summary>
        /// <returns>The next symbol, or "" as end marker</returns>
        ///---------------------------------------------------------------------
        public virtual string GetNextSymbol()
        {
            if (m_EnumerationListIndex < m_EnumerationList.Count)
               {
                   return m_EnumerationList[m_EnumerationListIndex++];
               }

            m_EnumerationList.Clear();
            m_EnumerationListIndex = 0;



            if (m_EnumerationCurrentLine < Lines.Count)
            {
                MatchCollection m = m_EnumerationRegex.Matches(Lines[m_EnumerationCurrentLine]);
                
                for (int i = 0; i < m.Count; i++)
                {
                    m_EnumerationList.Add(m[i].ToString());
                }
                m_EnumerationListIndex = 0;
                m_EnumerationCurrentLine++;

                if (m_EnumerationList.Count > 0)
                    return (m_EnumerationList[m_EnumerationListIndex++]).ToUpper();
                else
                    return "";
            }
            else
                return "";
        }




        public static string StringListToString(List<string> list)
        {
            StringBuilder sb = new StringBuilder("");
            for (int i = 0; i < list.Count; i++)
                if (i < list.Count - 1)
                    sb.AppendLine(list[i]);
                else
                    sb.Append(list[i]);
            return sb.ToString();
        }


        public static void StringToStringList(string lines, List<string> list)
        {
//            string[] theArray = Regex.Split(lines, System.Environment.NewLine);
            string[] theArray = Regex.Split(lines, @"\r\n|\n"); // zu-tun

            list.Clear();
            foreach (string s in theArray)
            {
                list.Add(s);
            }
        }



        public virtual string Text
        {
            get 
            {
                return StringListToString(Lines);
            }

            set 
            {
                StringToStringList(value, Lines);
            }
        }




        
    }







    ///-------------------------------------------------------------------------
    /// <summary>
    /// Represents a textfile. Nonempty subsequent lines make up a LineRecord.
    /// One or more empty lines separate these records.
    /// </summary>
    ///-------------------------------------------------------------------------
    public class LineRecordFile: IDocumentList, IClusterProvider
    {
        public List<LineRecord> Records = new List<LineRecord>();
        int       m_CurrentIDocumentList;    // For IDocumentList enumeration
        int       m_CurrentClusterProvider;  // For IClusterProvider enumeration
        Encoding  m_Encoding;

        public LineRecordFile(Encoding encoding)
        {
            m_Encoding = encoding;
        }



        public List<string> GetFirstCluster()
        {
            m_CurrentClusterProvider = 0;
            return GetNextCluster();
        }


        public List<string> GetNextCluster()
        {
            while (m_CurrentClusterProvider < Records.Count)
            {
                LineRecord lrec = Records[m_CurrentClusterProvider++];
                if (lrec.GetType() == typeof(LineRecord)) // IsA zu-tun
                    return lrec.Lines;
                else
                if (lrec.GetType() == typeof(LineRecordCluster))
                    return ((LineRecordCluster)lrec).Lines;
            }

            return null;
        }






        public void Clear()
        {
            Records.Clear();
        }

        protected void AddCluster(List<string> cluster)
        {
            if (cluster.Count <= 0)
                return;

            LineRecord record = new LineRecord(cluster.Count);
            record.Lines.AddRange(cluster);
            Records.Add(record);
            cluster.Clear();
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Appends line from a textfile to the internal ArrayList m_Records.
        /// If you want to clear the internal representation, use Clear().
        /// </summary>
        /// <param name="filename">The filename of a text file.</param>
        ///---------------------------------------------------------------------
        public void LoadFromFile(string filename)
        {
            List<string> cluster = new List<string>();
            Regex regexSpace = new Regex(@"^\s*$", RegexOptions.Compiled);

            StreamReader sr = null;

            FileStream fs = new FileStream(filename, FileMode.Open);

            BufferedStream bs = new BufferedStream(fs);
            
            sr = new StreamReader(bs, m_Encoding);
            try
            {
                string line = "";



                while ((line = sr.ReadLine()) != null)
                {
                    if (regexSpace.IsMatch(line))
                        AddCluster(cluster);
                    else
                        cluster.Add(line);
                }




                AddCluster(cluster);
            }
            finally
            {
                sr.Close();
            }
        }


        ///---------------------------------------------------------------------
        /// <summary>
        /// Does the same as LoadFromFile(string filename), but translates
        /// every record, if possible, into a more specialized subtype of LineRecord.
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="specializedTypes"></param>
        ///---------------------------------------------------------------------
        public void LoadFromFile(string filename, bool specializedTypes)
        {
            LoadFromFile(filename);
            
            if (!specializedTypes)
               return;

           LineRecordParser parser = new LineRecordParser();
           for (int i = 0; i < Records.Count; i++)
           {
               LineRecord oldObject = Records[i];
               LineRecord newObject = parser.GetSpecializedType(oldObject);
               Records[i] = newObject;
           }


        }



        ///---------------------------------------------------------------------
        /// <summary>
        /// Saves the file.
        /// </summary>
        ///---------------------------------------------------------------------
        public void SaveToFile(string filename)
        {
            StringBuilder sb = new StringBuilder("");
            string tempfile = Path.GetDirectoryName(filename) + @"\temp.$$$";

            for (int i = 0; i < Records.Count; i++)
            {
                LineRecord lrec = Records[i];
                if (!lrec.Deleted)
                   {
                    sb.AppendLine(lrec.Text);
                    sb.AppendLine();
                   }
            }


            FileStream stream = new FileStream(tempfile, FileMode.Create);
            if (!stream.CanWrite)
            {
                stream.Close();
                return;
            }

            StreamWriter writer = new StreamWriter(stream,m_Encoding);
            writer.Write(sb.ToString());
            writer.Flush();
            writer.Close();
            stream.Close();

            File.Delete(filename);
            File.Move(tempfile,filename);
        }



        public int Count
        {
            get
            {
                return Records.Count;
            }
        }


        public void Add(LineRecord lrec)
        {
            Records.Add(lrec);
        }

        public IDocument GetFirstDocument()
        {
            m_CurrentIDocumentList = -1;
            return GetNextDocument();
        }

        public IDocument GetNextDocument()
        {
            m_CurrentIDocumentList++;
            if (m_CurrentIDocumentList < Count)
                return Records[m_CurrentIDocumentList];
            else
                return null;
        }
    }
}
